"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for Computer Vision.

Example:
```python
import pixeltable as pxt
from pixeltable.functions import vision as pxtv

t = pxt.get_table(...)
t.select(pxtv.draw_bounding_boxes(t.img, boxes=t.boxes, label=t.labels)).collect()
```
"""

import colorsys
import hashlib
from collections import defaultdict
from typing import Any, Optional, Union

import numpy as np
import PIL.Image

import pixeltable as pxt
from pixeltable.utils.code import local_public_names


# the following function has been adapted from MMEval
# (sources at https://github.com/open-mmlab/mmeval)
# Copyright (c) OpenMMLab. All rights reserved.
def __calculate_bboxes_area(bboxes: np.ndarray) -> np.ndarray:
    """Calculate area of bounding boxes.

    Args:
        bboxes (numpy.ndarray): The bboxes with shape (n, 4) or (4, ) in 'xyxy' format.
     Returns:
        numpy.ndarray: The area of bboxes.
    """
    bboxes_w = bboxes[..., 2] - bboxes[..., 0]
    bboxes_h = bboxes[..., 3] - bboxes[..., 1]
    areas = bboxes_w * bboxes_h
    return areas


# the following function has been adapted from MMEval
# (sources at https://github.com/open-mmlab/mmeval)
# Copyright (c) OpenMMLab. All rights reserved.
def __calculate_overlaps(bboxes1: np.ndarray, bboxes2: np.ndarray) -> np.ndarray:
    """Calculate the overlap between each bbox of bboxes1 and bboxes2.

    Args:
        bboxes1 (numpy.ndarray): The bboxes with shape (n, 4) in 'xyxy' format.
        bboxes2 (numpy.ndarray): The bboxes with shape (k, 4) in 'xyxy' format.
    Returns:
        numpy.ndarray: IoUs or IoFs with shape (n, k).
    """
    bboxes1 = bboxes1.astype(np.float32)
    bboxes2 = bboxes2.astype(np.float32)
    rows = bboxes1.shape[0]
    cols = bboxes2.shape[0]
    overlaps = np.zeros((rows, cols), dtype=np.float32)

    if rows * cols == 0:
        return overlaps

    if bboxes1.shape[0] > bboxes2.shape[0]:
        # Swap bboxes for faster calculation.
        bboxes1, bboxes2 = bboxes2, bboxes1
        overlaps = np.zeros((cols, rows), dtype=np.float32)
        exchange = True
    else:
        exchange = False

    # Calculate the bboxes area.
    area1 = __calculate_bboxes_area(bboxes1)
    area2 = __calculate_bboxes_area(bboxes2)
    eps = np.finfo(np.float32).eps

    for i in range(bboxes1.shape[0]):
        x_start = np.maximum(bboxes1[i, 0], bboxes2[:, 0])
        y_start = np.maximum(bboxes1[i, 1], bboxes2[:, 1])
        x_end = np.minimum(bboxes1[i, 2], bboxes2[:, 2])
        y_end = np.minimum(bboxes1[i, 3], bboxes2[:, 3])
        overlap_w = np.maximum(x_end - x_start, 0)
        overlap_h = np.maximum(y_end - y_start, 0)
        overlap = overlap_w * overlap_h

        union = area1[i] + area2 - overlap
        union = np.maximum(union, eps)
        overlaps[i, :] = overlap / union
    return overlaps if not exchange else overlaps.T


# the following function has been adapted from MMEval
# (sources at https://github.com/open-mmlab/mmeval)
# Copyright (c) OpenMMLab. All rights reserved.
def __calculate_image_tpfp(
    pred_bboxes: np.ndarray, pred_scores: np.ndarray, gt_bboxes: np.ndarray, min_iou: float
) -> tuple[np.ndarray, np.ndarray]:
    """Calculate the true positive and false positive on an image.

    Args:
        pred_bboxes (numpy.ndarray): Predicted bboxes of this image, with
            shape (N, 5). The scores The predicted score of the bbox is
            concatenated behind the predicted bbox.
        gt_bboxes (numpy.ndarray): Ground truth bboxes of this image, with
            shape (M, 4).
        min_iou (float): The IoU threshold.

    Returns:
        tuple (tp, fp):
        tp: Shape (N,), the true positive flag of each predicted bbox on this image.
        fp: Shape (N,), the false positive flag of each predicted bbox on this image.
    """
    # Step 1. Concatenate `gt_bboxes` and `ignore_gt_bboxes`, then set
    # the `ignore_gt_flags`.
    # all_gt_bboxes = np.concatenate((gt_bboxes, ignore_gt_bboxes))
    # ignore_gt_flags = np.concatenate((np.zeros(
    #     (gt_bboxes.shape[0], 1),
    #     dtype=bool), np.ones((ignore_gt_bboxes.shape[0], 1), dtype=bool)))

    # Step 2. Initialize the `tp` and `fp` arrays.
    num_preds = pred_bboxes.shape[0]
    tp = np.zeros(num_preds, dtype=np.int8)
    fp = np.zeros(num_preds, dtype=np.int8)

    # Step 3. If there are no gt bboxes in this image, then all pred bboxes
    # within area range are false positives.
    if gt_bboxes.shape[0] == 0:
        fp[...] = 1
        return tp, fp

    # Step 4. Calculate the IoUs between the predicted bboxes and the
    # ground truth bboxes.
    ious = __calculate_overlaps(pred_bboxes, gt_bboxes)
    # For each pred bbox, the max iou with all gts.
    ious_max = ious.max(axis=1)
    # For each pred bbox, which gt overlaps most with it.
    ious_argmax = ious.argmax(axis=1)
    # Sort all pred bbox in descending order by scores.
    sorted_indices = np.argsort(-pred_scores)

    # Step 5. Count the `tp` and `fp` of each iou threshold and area range.
    # The flags that gt bboxes have been matched.
    gt_covered_flags = np.zeros(gt_bboxes.shape[0], dtype=bool)

    # Count the prediction bboxes in order of decreasing score.
    for pred_bbox_idx in sorted_indices:
        if ious_max[pred_bbox_idx] >= min_iou:
            matched_gt_idx = ious_argmax[pred_bbox_idx]
            if not gt_covered_flags[matched_gt_idx]:
                tp[pred_bbox_idx] = 1
                gt_covered_flags[matched_gt_idx] = True
            else:
                # This gt bbox has been matched and counted as fp.
                fp[pred_bbox_idx] = 1
        else:
            fp[pred_bbox_idx] = 1

    return tp, fp


@pxt.udf
def eval_detections(
    pred_bboxes: list[list[int]],
    pred_labels: list[int],
    pred_scores: list[float],
    gt_bboxes: list[list[int]],
    gt_labels: list[int],
    min_iou: float = 0.5,
) -> list[dict]:
    """
    Evaluates the performance of a set of predicted bounding boxes against a set of ground truth bounding boxes.

    Args:
        pred_bboxes: List of predicted bounding boxes, each represented as [xmin, ymin, xmax, ymax].
        pred_labels: List of predicted labels.
        pred_scores: List of predicted scores.
        gt_bboxes: List of ground truth bounding boxes, each represented as [xmin, ymin, xmax, ymax].
        gt_labels: List of ground truth labels.
        min_iou: Minimum intersection-over-union (IoU) threshold for a predicted bounding box to be
            considered a true positive.

    Returns:
        A list of dictionaries, one per label class, with the following structure:
        ```python
        {
            'min_iou': float,  # The value of `min_iou` used for the detections
            'class': int,  # The label class
            'tp': list[int],  # List of 1's and 0's indicating true positives for each
                              # predicted bounding box of this class
            'fp': list[int],  # List of 1's and 0's indicating false positives for each
                              # predicted bounding box of this class; `fp[n] == 1 - tp[n]`
            'scores': list[float],  # List of predicted scores for each bounding box of this class
            'num_gts': int,  # Number of ground truth bounding boxes of this class
        }
        ```
    """
    class_idxs = list(set(pred_labels + gt_labels))
    result: list[dict] = []
    pred_bboxes_arr = np.asarray(pred_bboxes)
    pred_classes_arr = np.asarray(pred_labels)
    pred_scores_arr = np.asarray(pred_scores)
    gt_bboxes_arr = np.asarray(gt_bboxes)
    gt_classes_arr = np.asarray(gt_labels)
    for class_idx in class_idxs:
        pred_filter = pred_classes_arr == class_idx
        gt_filter = gt_classes_arr == class_idx
        class_pred_scores = pred_scores_arr[pred_filter]
        tp, fp = __calculate_image_tpfp(
            pred_bboxes_arr[pred_filter], class_pred_scores, gt_bboxes_arr[gt_filter], min_iou
        )
        ordered_class_pred_scores = -np.sort(-class_pred_scores)
        result.append(
            {
                'min_iou': min_iou,
                'class': class_idx,
                'tp': tp.tolist(),
                'fp': fp.tolist(),
                'scores': ordered_class_pred_scores.tolist(),
                'num_gts': gt_filter.sum().item(),
            }
        )
    return result


@pxt.uda
class mean_ap(pxt.Aggregator):
    """
    Calculates the mean average precision (mAP) over
    [`eval_detections()`][pixeltable.functions.vision.eval_detections] results.

    __Parameters:__

    - `eval_dicts` (list[dict]): List of dictionaries as returned by
        [`eval_detections()`][pixeltable.functions.vision.eval_detections].

    __Returns:__

    - A `dict[int, float]` mapping each label class to an average precision (AP) value for that class.
    """

    def __init__(self) -> None:
        self.class_tpfp: dict[int, list[dict]] = defaultdict(list)

    def update(self, eval_dicts: list[dict]) -> None:
        for eval_dict in eval_dicts:
            class_idx = eval_dict['class']
            self.class_tpfp[class_idx].append(eval_dict)

    def value(self) -> dict:
        eps = np.finfo(np.float32).eps
        result: dict[int, float] = {}
        for class_idx, tpfp in self.class_tpfp.items():
            tp = np.concatenate([x['tp'] for x in tpfp], axis=0)
            fp = np.concatenate([x['fp'] for x in tpfp], axis=0)
            num_gts = np.sum([x['num_gts'] for x in tpfp])
            scores = np.concatenate([np.asarray(x['scores']) for x in tpfp])
            sorted_idxs = np.argsort(-scores)
            tp_cumsum = tp[sorted_idxs].cumsum()
            fp_cumsum = fp[sorted_idxs].cumsum()
            precision = tp_cumsum / np.maximum(tp_cumsum + fp_cumsum, eps)
            recall = tp_cumsum / np.maximum(num_gts, eps)

            mrec = np.hstack((0, recall, 1))
            mpre = np.hstack((0, precision, 0))
            for i in range(mpre.shape[0] - 1, 0, -1):
                mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])
            ind = np.where(mrec[1:] != mrec[:-1])[0]
            ap = np.sum((mrec[ind + 1] - mrec[ind]) * mpre[ind + 1])
            result[class_idx] = ap.item()
        return result


def __create_label_colors(labels: list[Any]) -> dict[Any, str]:
    """
    Create random colors for labels such that a particular label always gets the same color.

    Returns:
        dict mapping labels to colors
    """
    distinct_labels = set(labels)
    result: dict[Any, str] = {}
    for label in distinct_labels:
        # consistent hash for the label
        label_hash = int(hashlib.md5(str(label).encode()).hexdigest(), 16)
        hue = (label_hash % 360) / 360.0
        rgb = colorsys.hsv_to_rgb(hue, 0.7, 0.95)
        hex_color = '#{:02x}{:02x}{:02x}'.format(int(rgb[0] * 255), int(rgb[1] * 255), int(rgb[2] * 255))
        result[label] = hex_color
    return result


@pxt.udf
def draw_bounding_boxes(
    img: PIL.Image.Image,
    boxes: list[list[int]],
    labels: Optional[list[Any]] = None,
    color: Optional[str] = None,
    box_colors: Optional[list[str]] = None,
    fill: bool = False,
    width: int = 1,
    font: Optional[str] = None,
    font_size: Optional[int] = None,
) -> PIL.Image.Image:
    """
    Draws bounding boxes on the given image.

    Labels can be any type that supports `str()` and is hashable (e.g., strings, ints, etc.).

    Colors can be specified as common HTML color names (e.g., 'red') supported by PIL's
    [`ImageColor`](https://pillow.readthedocs.io/en/stable/reference/ImageColor.html#imagecolor-module) module or as
    RGB hex codes (e.g., '#FF0000').

    If no colors are specified, this function randomly assigns each label a specific color based on a hash of the label.

    Args:
        img: The image on which to draw the bounding boxes.
        boxes: List of bounding boxes, each represented as [xmin, ymin, xmax, ymax].
        labels: List of labels for each bounding box.
        color: Single color to be used for all bounding boxes and labels.
        box_colors: List of colors, one per bounding box.
        fill: Whether to fill the bounding boxes with color.
        width: Width of the bounding box borders.
        font: Name of a system font or path to a TrueType font file, as required by
            [`PIL.ImageFont.truetype()`](https://pillow.readthedocs.io/en/stable/reference/ImageFont.html#PIL.ImageFont.truetype).
            If `None`, uses the default provided by
            [`PIL.ImageFont.load_default()`](https://pillow.readthedocs.io/en/stable/reference/ImageFont.html#PIL.ImageFont.load_default).
        font_size: Size of the font used for labels in points. Only used in conjunction with non-`None` `font` argument.

    Returns:
        The image with bounding boxes drawn on it.
    """
    color_params = sum([color is not None, box_colors is not None])
    if color_params > 1:
        raise ValueError("Only one of 'color' or 'box_colors' can be set")

    # ensure the number of labels matches the number of boxes
    num_boxes = len(boxes)
    if labels is None:
        labels = [None] * num_boxes
    elif len(labels) != num_boxes:
        raise ValueError('Number of boxes and labels must match')

    if box_colors is not None:
        if len(box_colors) != num_boxes:
            raise ValueError('Number of boxes and box colors must match')
    elif color is not None:
        box_colors = [color] * num_boxes
    else:
        label_colors = __create_label_colors(labels)
        box_colors = [label_colors[label] for label in labels]

    from PIL import ImageColor, ImageDraw, ImageFont

    # set default font if not provided
    txt_font: Union[ImageFont.ImageFont, ImageFont.FreeTypeFont] = (
        ImageFont.load_default() if font is None else ImageFont.truetype(font=font, size=font_size or 10)
    )

    img_to_draw = img.copy()
    draw = ImageDraw.Draw(img_to_draw, 'RGBA' if fill else 'RGB')

    # Draw bounding boxes
    for i, bbox in enumerate(boxes):
        # determine color for the current box and label
        color = box_colors[i % len(box_colors)]

        if fill:
            rgb_color = ImageColor.getrgb(color)
            fill_color = (*rgb_color, 100)  # semi-transparent
            draw.rectangle(bbox, outline=color, width=width, fill=fill_color)  # type: ignore[arg-type]
        else:
            draw.rectangle(bbox, outline=color, width=width)  # type: ignore[arg-type]

    # Now draw labels separately, so they are not obscured by the boxes
    for bbox, label in zip(boxes, labels):
        if label is not None:
            label_str = str(label)
            _, _, text_width, text_height = draw.textbbox((0, 0), label_str, font=txt_font)
            if bbox[1] - text_height - 2 >= 0:
                # draw text above the box
                y = bbox[1] - text_height - 2
            else:
                y = bbox[3]
            if bbox[0] + text_width + 2 < img.width:
                x = bbox[0]
            else:
                x = img.width - text_width - 2
            draw.rectangle((x, y, x + text_width + 1, y + text_height + 1), fill='black')
            draw.text((x, y), label_str, fill='white', font=txt_font)

    return img_to_draw


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import logging
import sys
from typing import Any, AsyncIterator, Iterable, Optional, cast

from pixeltable import catalog, exceptions as excs, exprs

from .data_row_batch import DataRowBatch
from .exec_node import ExecNode

_logger = logging.getLogger('pixeltable')


class AggregationNode(ExecNode):
    """
    In-memory aggregation for UDAs.

    At the moment, this returns all results in a single DataRowBatch.
    """

    group_by: Optional[list[exprs.Expr]]
    input_exprs: list[exprs.Expr]
    agg_fn_eval_ctx: exprs.RowBuilder.EvalCtx
    agg_fn_calls: list[exprs.FunctionCall]
    output_batch: DataRowBatch

    def __init__(
        self,
        tbl: catalog.TableVersionHandle,
        row_builder: exprs.RowBuilder,
        group_by: Optional[list[exprs.Expr]],
        agg_fn_calls: list[exprs.FunctionCall],
        input_exprs: Iterable[exprs.Expr],
        input: ExecNode,
    ):
        output_exprs: list[exprs.Expr] = [] if group_by is None else list(group_by)
        output_exprs.extend(agg_fn_calls)
        super().__init__(row_builder, output_exprs, input_exprs, input)
        self.input = input
        self.group_by = group_by
        self.input_exprs = list(input_exprs)
        self.agg_fn_eval_ctx = row_builder.create_eval_ctx(agg_fn_calls, exclude=self.input_exprs)
        # we need to make sure to refer to the same exprs that RowBuilder.eval() will use
        self.agg_fn_calls = [cast(exprs.FunctionCall, e) for e in self.agg_fn_eval_ctx.target_exprs]
        # create output_batch here, rather than in __iter__(), so we don't need to remember tbl and row_builder
        self.output_batch = DataRowBatch(tbl, row_builder, 0)

    def _reset_agg_state(self, row_num: int) -> None:
        for fn_call in self.agg_fn_calls:
            try:
                fn_call.reset_agg()
            except Exception as exc:
                _, _, exc_tb = sys.exc_info()
                expr_msg = f'init() function of the aggregate {fn_call}'
                raise excs.ExprEvalError(fn_call, expr_msg, exc, exc_tb, [], row_num) from exc

    def _update_agg_state(self, row: exprs.DataRow, row_num: int) -> None:
        for fn_call in self.agg_fn_calls:
            try:
                fn_call.update(row)
            except Exception as exc:
                _, _, exc_tb = sys.exc_info()
                expr_msg = f'update() function of the aggregate {fn_call}'
                input_vals = [row[d.slot_idx] for d in fn_call.dependencies()]
                raise excs.ExprEvalError(fn_call, expr_msg, exc, exc_tb, input_vals, row_num) from exc

    async def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        prev_row: Optional[exprs.DataRow] = None
        current_group: Optional[list[Any]] = None  # the values of the group-by exprs
        num_input_rows = 0
        async for row_batch in self.input:
            num_input_rows += len(row_batch)
            for row in row_batch:
                group = [row[e.slot_idx] for e in self.group_by] if self.group_by is not None else None
                if current_group is None:
                    current_group = group
                    self._reset_agg_state(0)
                if group != current_group:
                    # we're entering a new group, emit a row for the previous one
                    self.row_builder.eval(prev_row, self.agg_fn_eval_ctx, profile=self.ctx.profile)
                    self.output_batch.add_row(prev_row)
                    current_group = group
                    self._reset_agg_state(0)
                self._update_agg_state(row, 0)
                prev_row = row
        if prev_row is not None:
            # emit the last group
            self.row_builder.eval(prev_row, self.agg_fn_eval_ctx, profile=self.ctx.profile)
            self.output_batch.add_row(prev_row)

        self.output_batch.flush_imgs(None, self.stored_img_cols, self.flushed_img_slots)
        _logger.debug(f'AggregateNode: consumed {num_input_rows} rows, returning {len(self.output_batch.rows)} rows')
        yield self.output_batch



from typing import Any, Optional

import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

from .column_ref import ColumnRef
from .data_row import DataRow
from .expr import Expr
from .literal import Literal
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class SimilarityExpr(Expr):
    def __init__(self, col_ref: ColumnRef, item: Any, idx_name: Optional[str] = None):
        super().__init__(ts.FloatType())
        item_expr = Expr.from_object(item)
        if item_expr is None or not (item_expr.col_type.is_string_type() or item_expr.col_type.is_image_type()):
            raise excs.Error(f'similarity(): requires a string or a PIL.Image.Image object, not a {type(item)}')
        assert item_expr.col_type.is_string_type() or item_expr.col_type.is_image_type()

        self.components = [col_ref, item_expr]

        from pixeltable import index

        # determine index to use
        idx_dict = ColumnRef.find_embedding_index(col_ref.col, idx_name, 'similarity')
        assert len(idx_dict) == 1
        self.idx_info = next(iter(idx_dict.values()))
        idx = self.idx_info.idx
        assert isinstance(idx, index.EmbeddingIndex)

        if item_expr.col_type.is_string_type() and idx.string_embed is None:
            raise excs.Error(
                f'Embedding index {self.idx_info.name!r} on column {self.idx_info.col.name!r} does not have a '
                f'string embedding and does not support string queries'
            )
        if item_expr.col_type.is_image_type() and idx.image_embed is None:
            raise excs.Error(
                f'Embedding index {self.idx_info.name!r} on column {self.idx_info.col.name!r} does not have an '
                f'image embedding and does not support image queries'
            )
        self.id = self._create_id()

    def __repr__(self) -> str:
        return f'{self.components[0]}.similarity({self.components[1]})'

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('idx_name', self.idx_info.name)]

    def default_column_name(self) -> str:
        return 'similarity'

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        if not isinstance(self.components[1], Literal):
            raise excs.Error('similarity(): requires a string or a PIL.Image.Image object, not an expression')
        item = self.components[1].val
        from pixeltable import index

        assert isinstance(self.idx_info.idx, index.EmbeddingIndex)
        return self.idx_info.idx.similarity_clause(self.idx_info.val_col, item)

    def as_order_by_clause(self, is_asc: bool) -> Optional[sql.ColumnElement]:
        if not isinstance(self.components[1], Literal):
            raise excs.Error('similarity(): requires a string or a PIL.Image.Image object, not an expression')
        item = self.components[1].val
        from pixeltable import index

        assert isinstance(self.idx_info.idx, index.EmbeddingIndex)
        return self.idx_info.idx.order_by_clause(self.idx_info.val_col, item, is_asc)

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        raise excs.Error('similarity(): cannot be used in a computed column')

    def _as_dict(self) -> dict:
        return {'idx_name': self.idx_info.name, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> 'SimilarityExpr':
        iname = d.get('idx_name')
        assert len(components) == 2
        assert isinstance(components[0], ColumnRef)
        return cls(components[0], components[1], idx_name=iname)



from __future__ import annotations

import dataclasses
import importlib
import logging
import sys
from typing import Optional
from uuid import UUID

import sqlalchemy as sql

from pixeltable import env, exceptions as excs, type_system as ts
from pixeltable.metadata import schema

from .function import Function

_logger = logging.getLogger('pixeltable')


class FunctionRegistry:
    """
    A central registry for all Functions. Handles interactions with the backing store.
    Function are loaded from the store on demand.
    """

    _instance: Optional[FunctionRegistry] = None

    @classmethod
    def get(cls) -> FunctionRegistry:
        if cls._instance is None:
            cls._instance = FunctionRegistry()
        return cls._instance

    def __init__(self) -> None:
        self.stored_fns_by_id: dict[UUID, Function] = {}
        self.module_fns: dict[str, Function] = {}  # fqn -> Function
        self.type_methods: dict[ts.ColumnType.Type, dict[str, Function]] = {}

    def clear_cache(self) -> None:
        """
        Useful during testing
        """
        self.stored_fns_by_id = {}

    # def register_std_modules(self) -> None:
    #     """Register all submodules of pixeltable.functions"""
    #     root = sys.modules['pixeltable.functions']
    #     self.register_submodules(root)
    #
    # def register_submodules(self, mod: types.ModuleType) -> None:
    #     # TODO: this doesn't work
    #     for name, submod in mod.__dict__.items():
    #         if isinstance(submod, types.ModuleType):
    #             self.register_module(submod)
    #             self.register_submodules(submod)
    #
    # def register_module(self) -> None:
    #     """Register all Functions in the caller module"""
    #     caller_path = get_caller_module_path()
    #     mod = importlib.import_module(caller_path)
    #     for name in dir(mod):
    #         obj = getattr(mod, name)
    #         if isinstance(obj, Function):
    #             fn_path = f'{caller_path}.{name}'  # fully-qualified name
    #             self.module_fns[fn_path] = obj

    def register_function(self, fqn: str, fn: Function) -> None:
        if fqn in self.module_fns:
            raise excs.Error(f'A UDF with that name already exists: {fqn}')
        self.module_fns[fqn] = fn
        if fn.is_method or fn.is_property:
            base_type = fn.signatures[0].parameters_by_pos[0].col_type.type_enum
            if base_type not in self.type_methods:
                self.type_methods[base_type] = {}
            if fn.name in self.type_methods[base_type]:
                raise excs.Error(f'Duplicate method name for type {base_type}: {fn.name}')
            self.type_methods[base_type][fn.name] = fn

    def list_functions(self) -> list[Function]:
        # retrieve Function.Metadata data for all existing stored functions from store directly
        # (self.stored_fns_by_id isn't guaranteed to contain all functions)
        # TODO: have the client do this, once the client takes over the Db functionality
        # stmt = sql.select(
        #         schema.Function.name, schema.Function.md,
        #         schema.Db.name, schema.Dir.path, sql_func.length(schema.Function.init_obj))\
        #     .where(schema.Function.db_id == schema.Db.id)\
        #     .where(schema.Function.dir_id == schema.Dir.id)
        # stored_fn_md: list[Function.Metadata] = []
        # with Env.get().engine.begin() as conn:
        #     rows = conn.execute(stmt)
        #     for name, md_dict, db_name, dir_path, init_obj_len in rows:
        #         md = Function.Metadata.from_dict(md_dict)
        #         md.fqn = f'{db_name}{"." + dir_path if dir_path != "" else ""}.{name}'
        #         stored_fn_md.append(md)
        return list(self.module_fns.values())

    # def get_function(self, *, id: Optional[UUID] = None, fqn: Optional[str] = None) -> Function:
    #     assert (id is not None) != (fqn is not None)
    #     if id is not None:
    #         if id not in self.stored_fns_by_id:
    #             stmt = sql.select(
    #                     schema.Function.md, schema.Function.eval_obj, schema.Function.init_obj,
    #                     schema.Function.update_obj, schema.Function.value_obj) \
    #                 .where(schema.Function.id == id)
    #             with env.Env.get().engine.begin() as conn:
    #                 rows = conn.execute(stmt)
    #                 row = next(rows)
    #                 schema_md = schema.md_from_dict(schema.FunctionMd, row[0])
    #                 name = schema_md.name
    #                 md = FunctionMd.from_dict(schema_md.md)
    #                 # md.fqn is set by caller
    #                 eval_fn = cloudpickle.loads(row[1]) if row[1] is not None else None
    #                 # TODO: are these checks needed?
    #                 if row[1] is not None and eval_fn is None:
    #                     raise excs.Error(f'Could not load eval_fn for function {name}')
    #                 init_fn = cloudpickle.loads(row[2]) if row[2] is not None else None
    #                 if row[2] is not None and init_fn is None:
    #                     raise excs.Error(f'Could not load init_fn for aggregate function {name}')
    #                 update_fn = cloudpickle.loads(row[3]) if row[3] is not None else None
    #                 if row[3] is not None and update_fn is None:
    #                     raise excs.Error(f'Could not load update_fn for aggregate function {name}')
    #                 value_fn = cloudpickle.loads(row[4]) if row[4] is not None else None
    #                 if row[4] is not None and value_fn is None:
    #                     raise excs.Error(f'Could not load value_fn for aggregate function {name}')
    #
    #                 func = Function(
    #                     md, id=id,
    #                     eval_fn=eval_fn, init_fn=init_fn, update_fn=update_fn, value_fn=value_fn)
    #                 _logger.info(f'Loaded function {name} from store')
    #                 self.stored_fns_by_id[id] = func
    #         assert id in self.stored_fns_by_id
    #         return self.stored_fns_by_id[id]
    #     else:
    #         # this is an already-registered library function
    #         assert fqn in self.module_fns, f'{fqn} not found'
    #         return self.module_fns[fqn]

    def get_type_methods(self, base_type: ts.ColumnType.Type) -> list[Function]:
        """
        Get a list of all methods (and properties) registered for a given base type.
        """
        if base_type in self.type_methods:
            return list(self.type_methods[base_type].values())
        return []

    def lookup_type_method(self, base_type: ts.ColumnType.Type, name: str) -> Optional[Function]:
        """
        Look up a method (or property) by name for a given base type. If no such method is registered, return None.
        """
        if base_type in self.type_methods and name in self.type_methods[base_type]:
            return self.type_methods[base_type][name]
        return None

    # def create_function(self, md: schema.FunctionMd, binary_obj: bytes, dir_id: Optional[UUID] = None) -> UUID:
    def create_stored_function(self, pxt_fn: Function, dir_id: Optional[UUID] = None) -> UUID:
        fn_md, binary_obj = pxt_fn.to_store()
        md = schema.FunctionMd(name=pxt_fn.name, md=fn_md, py_version=sys.version, class_name=pxt_fn.__class__.__name__)
        with env.Env.get().engine.begin() as conn:
            res = conn.execute(
                sql.insert(schema.Function.__table__).values(
                    dir_id=dir_id, md=dataclasses.asdict(md), binary_obj=binary_obj
                )
            )
            id = res.inserted_primary_key[0]
            _logger.info(f'Created function {pxt_fn.name} (id {id}) in store')
            self.stored_fns_by_id[id] = pxt_fn
            return id

    def get_stored_function(self, id: UUID) -> Function:
        if id in self.stored_fns_by_id:
            return self.stored_fns_by_id[id]
        stmt = sql.select(schema.Function.md, schema.Function.binary_obj, schema.Function.dir_id).where(
            schema.Function.id == id
        )
        with env.Env.get().engine.begin() as conn:
            row = conn.execute(stmt).fetchone()
            if row is None:
                raise excs.Error(f'Function with id {id} not found')
            # create instance of the referenced class
            md = schema.md_from_dict(schema.FunctionMd, row[0])
            func_module = importlib.import_module(self.__module__.rsplit('.', 1)[0])
            func_class = getattr(func_module, md.class_name)
            instance = func_class.from_store(md.name, md.md, row[1])
            self.stored_fns_by_id[id] = instance
            return instance

    # def create_function(self, fn: Function, dir_id: Optional[UUID] = None, name: Optional[str] = None) -> None:
    #     with env.Env.get().engine.begin() as conn:
    #         _logger.debug(f'Pickling function {name}')
    #         eval_fn_str = cloudpickle.dumps(fn.eval_fn) if fn.eval_fn is not None else None
    #         init_fn_str = cloudpickle.dumps(fn.init_fn) if fn.init_fn is not None else None
    #         update_fn_str = cloudpickle.dumps(fn.update_fn) if fn.update_fn is not None else None
    #         value_fn_str = cloudpickle.dumps(fn.value_fn) if fn.value_fn is not None else None
    #         total_size = \
    #             (len(eval_fn_str) if eval_fn_str is not None else 0) + \
    #             (len(init_fn_str) if init_fn_str is not None else 0) + \
    #             (len(update_fn_str) if update_fn_str is not None else 0) + \
    #             (len(value_fn_str) if value_fn_str is not None else 0)
    #         _logger.debug(f'Pickled function {name} ({total_size} bytes)')
    #
    #         schema_md = schema.FunctionMd(name=name, md=fn.md.as_dict())
    #         res = conn.execute(
    #             sql.insert(schema.Function.__table__)
    #                 .values(
    #                     dir_id=dir_id, md=dataclasses.asdict(schema_md),
    #                     eval_obj=eval_fn_str, init_obj=init_fn_str, update_obj=update_fn_str, value_obj=value_fn_str))
    #         fn.id = res.inserted_primary_key[0]
    #         self.stored_fns_by_id[fn.id] = fn
    #         _logger.info(f'Created function {name} in store')

    # def update_function(self, id: UUID, new_fn: Function) -> None:
    #     """
    #     Updates the callables for the Function with the given id in the store and in the cache, if present.
    #     """
    #     assert not new_fn.is_module_function
    #     with env.Env.get().engine.begin() as conn:
    #         updates = {}
    #         if new_fn.eval_fn is not None:
    #             updates[schema.Function.eval_obj] = cloudpickle.dumps(new_fn.eval_fn)
    #         if new_fn.init_fn is not None:
    #             updates[schema.Function.init_obj] = cloudpickle.dumps(new_fn.init_fn)
    #         if new_fn.update_fn is not None:
    #             updates[schema.Function.update_obj] = cloudpickle.dumps(new_fn.update_fn)
    #         if new_fn.value_fn is not None:
    #             updates[schema.Function.value_obj] = cloudpickle.dumps(new_fn.value_fn)
    #         conn.execute(
    #             sql.update(schema.Function.__table__)
    #                 .values(updates)
    #                 .where(schema.Function.id == id))
    #         _logger.info(f'Updated function {new_fn.md.fqn} (id={id}) in store')
    #     if id in self.stored_fns_by_id:
    #         if new_fn.eval_fn is not None:
    #             self.stored_fns_by_id[id].eval_fn = new_fn.eval_fn
    #         if new_fn.init_fn is not None:
    #             self.stored_fns_by_id[id].init_fn = new_fn.init_fn
    #         if new_fn.update_fn is not None:
    #             self.stored_fns_by_id[id].update_fn = new_fn.update_fn
    #         if new_fn.value_fn is not None:
    #             self.stored_fns_by_id[id].value_fn = new_fn.value_fn

    def delete_function(self, id: UUID) -> None:
        assert id is not None
        with env.Env.get().engine.begin() as conn:
            conn.execute(sql.delete(schema.Function.__table__).where(schema.Function.id == id))
            _logger.info(f'Deleted function with id {id} from store')



import dataclasses
import io
import json
import logging
import tarfile
import urllib.parse
import urllib.request
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator, Optional

import more_itertools
import numpy as np
import pyarrow as pa
import pyiceberg.catalog

import pixeltable as pxt
import pixeltable.type_system as ts
from pixeltable import catalog, exprs, metadata
from pixeltable.dataframe import DataFrame
from pixeltable.env import Env
from pixeltable.utils.arrow import PXT_TO_PA_TYPES
from pixeltable.utils.iceberg import sqlite_catalog

_logger = logging.getLogger('pixeltable')


class TablePackager:
    """
    Packages a pixeltable Table into a tarball containing Iceberg tables and media files. The structure of the tarball
    is as follows:

    metadata.json  # Pixeltable metadata for the packaged table
    warehouse/catalog.db  # sqlite Iceberg catalog
    warehouse/pxt.db/**  # Iceberg metadata and data files (parquet/avro/json)
    media/**  # Local media files

    If the table being archived is a view, then the Iceberg catalog will contain separate tables for the view and each
    of its ancestors. All rows will be exported with additional _rowid and _v_min columns. Currently, only the most
    recent version of the table can be exported, and only the full table contents.

    If the table contains media columns, they are handled as follows:
    - If a media file has an external URL (any URL scheme other than file://), then the URL will be preserved as-is and
      stored in the Iceberg table.
    - If a media file is a local file, then it will be copied into the tarball as a file of the form
      'media/{uuid}{extension}', and the Iceberg table will contain the ephemeral URI 'pxtmedia://{uuid}{extension}'.
    """

    table: catalog.Table  # The table to be packaged
    tmp_dir: Path  # Temporary directory where the package will reside
    iceberg_catalog: pyiceberg.catalog.Catalog
    media_files: dict[Path, str]  # Mapping from local media file paths to their tarball names
    md: dict[str, Any]

    def __init__(self, table: catalog.Table, additional_md: Optional[dict[str, Any]] = None) -> None:
        self.table = table
        self.tmp_dir = Path(Env.get().create_tmp_path())
        self.media_files = {}

        # Generate metadata
        self.md = {
            'pxt_version': pxt.__version__,
            'pxt_md_version': metadata.VERSION,
            'md': {
                'tables': [
                    {
                        'table_id': str(t._tbl_version.id),
                        # These are temporary; will replace with a better solution once the concurrency
                        # changes to catalog have been merged
                        'table_md': dataclasses.asdict(t._tbl_version.get()._create_tbl_md()),
                        'table_version_md': dataclasses.asdict(
                            t._tbl_version.get()._create_version_md(datetime.now().timestamp())
                        ),
                        'table_schema_version_md': dataclasses.asdict(
                            t._tbl_version.get()._create_schema_version_md(0)
                        ),
                    }
                    for t in (table, *table._bases)
                ]
            },
        }
        if additional_md is not None:
            self.md.update(additional_md)

    def package(self) -> Path:
        """
        Export the table to a tarball containing Iceberg tables and media files.
        """
        assert not self.tmp_dir.exists()  # Packaging can only be done once per TablePackager instance
        _logger.info(f"Packaging table '{self.table._path}' and its ancestors in: {self.tmp_dir}")
        self.tmp_dir.mkdir()
        with open(self.tmp_dir / 'metadata.json', 'w', encoding='utf8') as fp:
            json.dump(self.md, fp)
        self.iceberg_catalog = sqlite_catalog(self.tmp_dir / 'warehouse')
        with Env.get().begin_xact():
            ancestors = (self.table, *self.table._bases)
            for t in ancestors:
                _logger.info(f"Exporting table '{t._path}'.")
                self.__export_table(t)
        _logger.info('Building archive.')
        bundle_path = self.__build_tarball()
        _logger.info(f'Packaging complete: {bundle_path}')
        return bundle_path

    def __export_table(self, t: catalog.Table) -> None:
        """
        Exports the data from `t` into an Iceberg table.
        """
        # First generate a select list for the data we want to extract from `t`. This includes:
        # - all stored columns, including computed columns;
        # - errortype and errormsg fields whenever they're defined.
        # We select only those columns that are defined in this table (columns inherited from ancestor tables will be
        # handled separately).
        # For media columns, we substitute `col.fileurl` so that we always get the URL (which may be a file:// URL;
        # these will be specially handled later)
        select_exprs: dict[str, exprs.Expr] = {}

        # As we generate the select list, we construct a separate list of column types. We can't rely on df._schema
        # to get the column types, since we'll be substituting `fileurl`s for media columns.
        actual_col_types: list[ts.ColumnType] = []

        for col_name, col in t._tbl_version.get().cols_by_name.items():
            if not col.is_stored:
                continue
            if col.col_type.is_media_type():
                select_exprs[col_name] = t[col_name].fileurl
            else:
                select_exprs[col_name] = t[col_name]
            actual_col_types.append(col.col_type)
            if col.records_errors:
                select_exprs[f'{col_name}_errortype'] = t[col_name].errortype
                actual_col_types.append(ts.StringType())
                select_exprs[f'{col_name}_errormsg'] = t[col_name].errormsg
                actual_col_types.append(ts.StringType())

        # Run the select() on `self.table`, not `t`, so that we export only those rows that are actually present in
        # `self.table`.
        df = self.table.select(**select_exprs)
        namespace = self.__iceberg_namespace(t)
        self.iceberg_catalog.create_namespace_if_not_exists(namespace)
        iceberg_schema = self.__to_iceberg_schema(df._schema)
        iceberg_tbl = self.iceberg_catalog.create_table(f'{namespace}.{t._name}', schema=iceberg_schema)

        # Populate the Iceberg table with data.
        # The data is first loaded from the DataFrame into a sequence of pyarrow tables, batched in order to avoid
        # excessive memory usage. The pyarrow tables are then amalgamated into the (single) Iceberg table on disk.
        for pa_table in self.__to_pa_tables(df, actual_col_types, iceberg_schema):
            iceberg_tbl.append(pa_table)

    @classmethod
    def __iceberg_namespace(cls, table: catalog.Table) -> str:
        """
        Iceberg tables must have a namespace, which cannot be the empty string, so we prepend `pxt` to the table path.
        """
        parent_path = table._parent()._path()
        if len(parent_path) == 0:
            return 'pxt'
        else:
            return f'pxt.{parent_path}'

    # The following methods are responsible for schema and data conversion from Pixeltable to Iceberg. Some of this
    # logic might be consolidated into arrow.py and unified with general Parquet export, but there are several
    # major differences:
    # - Iceberg has no array type; we export all arrays as binary blobs
    # - We include _rowid and _v_min columns in the Iceberg table
    # - Media columns are handled specially as indicated above

    @classmethod
    def __to_iceberg_schema(cls, pxt_schema: dict[str, ts.ColumnType]) -> pa.Schema:
        entries = [(name, cls.__to_iceberg_type(col_type)) for name, col_type in pxt_schema.items()]
        entries.append(('_rowid', pa.list_(pa.int64())))
        entries.append(('_v_min', pa.int64()))
        return pa.schema(entries)  # type: ignore[arg-type]

    @classmethod
    def __to_iceberg_type(cls, col_type: ts.ColumnType) -> pa.DataType:
        if col_type.is_array_type():
            return pa.binary()
        if col_type.is_media_type():
            return pa.string()
        return PXT_TO_PA_TYPES.get(col_type.__class__)

    def __to_pa_tables(
        self, df: DataFrame, actual_col_types: list[ts.ColumnType], arrow_schema: pa.Schema, batch_size: int = 1_000
    ) -> Iterator[pa.Table]:
        """
        Load a DataFrame as a sequence of pyarrow tables. The pyarrow tables are batched into smaller chunks
        to avoid excessive memory usage.
        """
        for rows in more_itertools.batched(self.__to_pa_rows(df, actual_col_types), batch_size):
            cols = {col_name: [row[idx] for row in rows] for idx, col_name in enumerate(df._schema.keys())}
            cols['_rowid'] = [row[-2] for row in rows]
            cols['_v_min'] = [row[-1] for row in rows]
            yield pa.Table.from_pydict(cols, schema=arrow_schema)

    def __to_pa_rows(self, df: DataFrame, actual_col_types: list[ts.ColumnType]) -> Iterator[list]:
        for row in df._exec():
            vals = [row[e.slot_idx] for e in df._select_list_exprs]
            result = [self.__to_pa_value(val, col_type) for val, col_type in zip(vals, actual_col_types)]
            result.append(row.rowid)
            result.append(row.v_min)
            yield result

    def __to_pa_value(self, val: Any, col_type: ts.ColumnType) -> Any:
        if val is None:
            return None
        if col_type.is_array_type():
            # Export arrays as binary
            assert isinstance(val, np.ndarray)
            arr = io.BytesIO()
            np.save(arr, val)
            return arr.getvalue()
        if col_type.is_json_type():
            # Export JSON as strings
            return json.dumps(val)
        if col_type.is_media_type():
            # Handle media files as described above
            assert isinstance(val, str)  # Media columns are always referenced by `fileurl`
            return self.__process_media_url(val)
        return val

    def __process_media_url(self, url: str) -> str:
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme == 'file':
            # It's the URL of a local file. Replace it with a pxtmedia:// URI.
            # (We can't use an actual pxt:// URI, because the eventual pxt:// table name might not be known at this
            # time. The pxtmedia:// URI serves as a relative reference into the tarball that can be replaced with an
            # actual URL when the table is reconstituted.)
            path = Path(urllib.parse.unquote(urllib.request.url2pathname(parsed_url.path)))
            if path not in self.media_files:
                # Create a new entry in the `media_files` dict so that we can copy the file into the tarball later.
                dest_name = f'{uuid.uuid4().hex}{path.suffix}'
                self.media_files[path] = dest_name
            return f'pxtmedia://{self.media_files[path]}'
        # For any type of URL other than a local file, just return the URL as-is.
        return url

    def __build_tarball(self) -> Path:
        bundle_path = self.tmp_dir / 'bundle.tar.bz2'
        with tarfile.open(bundle_path, 'w:bz2') as tf:
            # Add metadata json
            tf.add(self.tmp_dir / 'metadata.json', arcname='metadata.json')
            # Add the Iceberg warehouse dir (including the catalog)
            tf.add(self.tmp_dir / 'warehouse', arcname='warehouse', recursive=True)
            # Add the media files
            for src_file, dest_name in self.media_files.items():
                tf.add(src_file, arcname=f'media/{dest_name}')
        return bundle_path



from __future__ import annotations

from typing import Any, Optional, Union

import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

from .data_row import DataRow
from .expr import Expr
from .globals import StringOperator
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class StringOp(Expr):
    """
    Allows operations on strings
    """

    operator: StringOperator

    def __init__(self, operator: StringOperator, op1: Expr, op2: Expr):
        super().__init__(ts.StringType(nullable=op1.col_type.nullable))
        self.operator = operator
        self.components = [op1, op2]
        assert op1.col_type.is_string_type()
        if operator in (StringOperator.CONCAT, StringOperator.REPEAT):
            if operator == StringOperator.CONCAT and not op2.col_type.is_string_type():
                raise excs.Error(
                    f'{self}: {operator} on strings requires string type, but {op2} has type {op2.col_type}'
                )
            if operator == StringOperator.REPEAT and not op2.col_type.is_int_type():
                raise excs.Error(f'{self}: {operator} on strings requires int type, but {op2} has type {op2.col_type}')
        else:
            raise excs.Error(
                f'{self}: invalid operation {operator} on strings; '
                f'only operators {StringOperator.CONCAT} and {StringOperator.REPEAT} are supported'
            )
        self.id = self._create_id()

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def __repr__(self) -> str:
        # add parentheses around operands that are StringOpExpr to express precedence
        op1_str = f'({self._op1})' if isinstance(self._op1, StringOp) else str(self._op1)
        op2_str = f'({self._op2})' if isinstance(self._op2, StringOp) else str(self._op2)
        return f'{op1_str} {self.operator} {op2_str}'

    def _equals(self, other: StringOp) -> bool:
        return self.operator == other.operator

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('operator', self.operator.value)]

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        left = sql_elements.get(self._op1)
        right = sql_elements.get(self._op2)
        if left is None or right is None:
            return None
        if self.operator == StringOperator.CONCAT:
            return left.concat(right)
        if self.operator == StringOperator.REPEAT:
            return sql.func.repeat(sql.cast(left, sql.String), sql.cast(right, sql.Integer))
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        op1_val = data_row[self._op1.slot_idx]
        op2_val = data_row[self._op2.slot_idx]
        data_row[self.slot_idx] = self.eval_nullable(op1_val, op2_val)

    def eval_nullable(self, op1_val: Union[str, None], op2_val: Union[int, str, None]) -> Union[str, None]:
        """
        Return the result of evaluating the expression on two nullable int/float operands,
        None is interpreted as SQL NULL
        """
        if op1_val is None or op2_val is None:
            return None
        return self.eval_non_null(op1_val, op2_val)

    def eval_non_null(self, op1_val: str, op2_val: Union[int, str]) -> str:
        """
        Return the result of evaluating the expression on two int/float operands
        """
        assert self.operator in (StringOperator.CONCAT, StringOperator.REPEAT)
        if self.operator == StringOperator.CONCAT:
            assert isinstance(op2_val, str)
            return op1_val + op2_val
        else:
            assert isinstance(op2_val, int)
            return op1_val * op2_val

    def _as_dict(self) -> dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> StringOp:
        assert 'operator' in d
        assert len(components) == 2
        return cls(StringOperator(d['operator']), components[0], components[1])



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, Optional, overload

from pixeltable import exprs, type_system as ts

from .function import Function
from .signature import Signature

if TYPE_CHECKING:
    from pixeltable import DataFrame


class QueryTemplateFunction(Function):
    """A parameterized query/DataFrame from which an executable DataFrame is created with a function call."""

    template_df: Optional['DataFrame']
    self_name: Optional[str]

    @classmethod
    def create(
        cls, template_callable: Callable, param_types: Optional[list[ts.ColumnType]], path: str, name: str
    ) -> QueryTemplateFunction:
        # we need to construct a template df and a signature
        py_sig = inspect.signature(template_callable)
        py_params = list(py_sig.parameters.values())
        params = Signature.create_parameters(py_params=py_params, param_types=param_types)
        # invoke template_callable with parameter expressions to construct a DataFrame with parameters
        var_exprs = [exprs.Variable(param.name, param.col_type) for param in params]
        template_df = template_callable(*var_exprs)
        from pixeltable import DataFrame

        assert isinstance(template_df, DataFrame)
        # we take params and return json
        sig = Signature(return_type=ts.JsonType(), parameters=params)
        return QueryTemplateFunction(template_df, sig, path=path, name=name)

    def __init__(
        self, template_df: Optional['DataFrame'], sig: Signature, path: Optional[str] = None, name: Optional[str] = None
    ):
        assert sig is not None
        super().__init__([sig], self_path=path)
        self.self_name = name
        self.template_df = template_df

    def _update_as_overload_resolution(self, signature_idx: int) -> None:
        pass  # only one signature supported for QueryTemplateFunction

    @property
    def is_async(self) -> bool:
        return True

    async def aexec(self, *args: Any, **kwargs: Any) -> Any:
        # assert not self.is_polymorphic
        bound_args = self.signature.py_signature.bind(*args, **kwargs).arguments
        # apply defaults, otherwise we might have Parameters left over
        bound_args.update(
            {
                param.name: param.default
                for param in self.signature.parameters.values()
                if param.has_default() and param.name not in bound_args
            }
        )
        bound_df = self.template_df.bind(bound_args)
        result = await bound_df._acollect()
        return list(result)

    @property
    def display_name(self) -> str:
        return self.self_name

    @property
    def name(self) -> str:
        return self.self_name

    def _as_dict(self) -> dict:
        return {'name': self.name, 'signature': self.signature.as_dict(), 'df': self.template_df.as_dict()}

    @classmethod
    def _from_dict(cls, d: dict) -> Function:
        from pixeltable.dataframe import DataFrame

        return cls(DataFrame.from_dict(d['df']), Signature.from_dict(d['signature']), name=d['name'])


@overload
def query(py_fn: Callable) -> QueryTemplateFunction: ...


@overload
def query(*, param_types: Optional[list[ts.ColumnType]] = None) -> Callable[[Callable], QueryTemplateFunction]: ...


def query(*args: Any, **kwargs: Any) -> Any:
    def make_query_template(py_fn: Callable, param_types: Optional[list[ts.ColumnType]]) -> QueryTemplateFunction:
        if py_fn.__module__ != '__main__' and py_fn.__name__.isidentifier():
            # this is a named function in a module
            function_path = f'{py_fn.__module__}.{py_fn.__qualname__}'
        else:
            function_path = None
        query_name = py_fn.__name__
        query_fn = QueryTemplateFunction.create(py_fn, param_types=param_types, path=function_path, name=query_name)
        return query_fn

        # TODO: verify that the inferred return type matches that of the template
        # TODO: verify that the signature doesn't contain batched parameters

    if len(args) == 1:
        assert len(kwargs) == 0 and callable(args[0])
        return make_query_template(args[0], None)
    else:
        assert len(args) == 0 and len(kwargs) == 1 and 'param_types' in kwargs
        return lambda py_fn: make_query_template(py_fn, kwargs['param_types'])



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `ImageType`.

Example:
```python
import pixeltable as pxt

t = pxt.get_table(...)
t.select(t.img_col.convert('L')).collect()
```
"""

import base64
from typing import Optional

import PIL.Image

import pixeltable as pxt
from pixeltable.exprs import Expr
from pixeltable.utils.code import local_public_names


@pxt.udf(is_method=True)
def b64_encode(img: PIL.Image.Image, image_format: str = 'png') -> str:
    """
    Convert image to a base64-encoded string.

    Args:
        img: image
        image_format: image format [supported by PIL](https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#fully-supported-formats)
    """
    import io

    bytes_arr = io.BytesIO()
    img.save(bytes_arr, format=image_format)
    b64_bytes = base64.b64encode(bytes_arr.getvalue())
    return b64_bytes.decode('utf-8')


@pxt.udf(substitute_fn=PIL.Image.alpha_composite, is_method=True)
def alpha_composite(im1: PIL.Image.Image, im2: PIL.Image.Image) -> PIL.Image.Image:
    """
    Alpha composite `im2` over `im1`.

    Equivalent to [`PIL.Image.alpha_composite()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.alpha_composite)
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.blend, is_method=True)
def blend(im1: PIL.Image.Image, im2: PIL.Image.Image, alpha: float) -> PIL.Image.Image:
    """
    Return a new image by interpolating between two input images, using a constant alpha.

    Equivalent to [`PIL.Image.blend()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.blend)
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.composite, is_method=True)
def composite(image1: PIL.Image.Image, image2: PIL.Image.Image, mask: PIL.Image.Image) -> PIL.Image.Image:
    """
    Return a composite image by blending two images using a mask.

    Equivalent to [`PIL.Image.composite()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.composite)
    """
    pass


# PIL.Image.Image methods


# Image.convert()
@pxt.udf(is_method=True)
def convert(self: PIL.Image.Image, mode: str) -> PIL.Image.Image:
    """
    Convert the image to a different mode.

    Equivalent to
    [`PIL.Image.Image.convert()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.convert).

    Args:
        mode: The mode to convert to. See the
            [Pillow documentation](https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes)
            for a list of supported modes.
    """
    return self.convert(mode)


@convert.conditional_return_type
def _(self: Expr, mode: str) -> pxt.ColumnType:
    input_type = self.col_type
    assert isinstance(input_type, pxt.ImageType)
    return pxt.ImageType(size=input_type.size, mode=mode, nullable=input_type.nullable)


# Image.crop()
@pxt.udf(substitute_fn=PIL.Image.Image.crop, is_method=True)
def crop(self: PIL.Image.Image, box: tuple[int, int, int, int]) -> PIL.Image.Image:
    """
    Return a rectangular region from the image. The box is a 4-tuple defining the left, upper, right, and lower pixel
    coordinates.

    Equivalent to
    [`PIL.Image.Image.crop()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.crop)
    """
    pass


@crop.conditional_return_type
def _(self: Expr, box: tuple[int, int, int, int]) -> pxt.ColumnType:
    input_type = self.col_type
    assert isinstance(input_type, pxt.ImageType)
    if (isinstance(box, (list, tuple))) and len(box) == 4 and all(isinstance(x, int) for x in box):
        return pxt.ImageType(
            size=(box[2] - box[0], box[3] - box[1]), mode=input_type.mode, nullable=input_type.nullable
        )
    return pxt.ImageType(mode=input_type.mode, nullable=input_type.nullable)  # we can't compute the size statically


# Image.getchannel()
@pxt.udf(substitute_fn=PIL.Image.Image.getchannel, is_method=True)
def getchannel(self: PIL.Image.Image, channel: int) -> PIL.Image.Image:
    """
    Return an L-mode image containing a single channel of the original image.

    Equivalent to
    [`PIL.Image.Image.getchannel()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getchannel)

    Args:
        channel: The channel to extract. This is a 0-based index.
    """
    pass


@getchannel.conditional_return_type
def _(self: Expr) -> pxt.ColumnType:
    input_type = self.col_type
    assert isinstance(input_type, pxt.ImageType)
    return pxt.ImageType(size=input_type.size, mode='L', nullable=input_type.nullable)


@pxt.udf(is_method=True)
def get_metadata(self: PIL.Image.Image) -> dict:
    """
    Return metadata for the image.
    """
    return {
        'width': self.width,
        'height': self.height,
        'mode': self.mode,
        'bits': getattr(self, 'bits', None),
        'format': self.format,
        'palette': self.palette,
    }


# Image.point()
@pxt.udf(is_method=True)
def point(self: PIL.Image.Image, lut: list[int], mode: Optional[str] = None) -> PIL.Image.Image:
    """
    Map image pixels through a lookup table.

    Equivalent to
    [`PIL.Image.Image.point()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.point)

    Args:
        lut: A lookup table.
    """
    return self.point(lut, mode=mode)


# Image.resize()
@pxt.udf(is_method=True)
def resize(self: PIL.Image.Image, size: tuple[int, int]) -> PIL.Image.Image:
    """
    Return a resized copy of the image. The size parameter is a tuple containing the width and height of the new image.

    Equivalent to
    [`PIL.Image.Image.resize()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.resize)
    """
    return self.resize(tuple(size))  # type: ignore[arg-type]


@resize.conditional_return_type
def _(self: Expr, size: tuple[int, int]) -> pxt.ColumnType:
    input_type = self.col_type
    assert isinstance(input_type, pxt.ImageType)
    return pxt.ImageType(size=size, mode=input_type.mode, nullable=input_type.nullable)


# Image.rotate()
@pxt.udf(is_method=True)
def rotate(self: PIL.Image.Image, angle: int) -> PIL.Image.Image:
    """
    Return a copy of the image rotated by the given angle.

    Equivalent to
    [`PIL.Image.Image.rotate()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.rotate)

    Args:
        angle: The angle to rotate the image, in degrees. Positive angles are counter-clockwise.
    """
    return self.rotate(angle)


@pxt.udf(substitute_fn=PIL.Image.Image.effect_spread, is_method=True)
def effect_spread(self: PIL.Image.Image, distance: int) -> PIL.Image.Image:
    """
    Randomly spread pixels in an image.

    Equivalent to
    [`PIL.Image.Image.effect_spread()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.effect_spread)

    Args:
        distance: The distance to spread pixels.
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.transpose, is_method=True)
def transpose(self: PIL.Image.Image, method: int) -> PIL.Image.Image:
    """
    Transpose the image.

    Equivalent to
    [`PIL.Image.Image.transpose()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.transpose)

    Args:
        method: The transpose method. See the
            [Pillow documentation](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.transpose)
            for a list of supported methods.
    """
    pass


@rotate.conditional_return_type
@effect_spread.conditional_return_type
@transpose.conditional_return_type
def _(self: Expr) -> pxt.ColumnType:
    return self.col_type


@pxt.udf(substitute_fn=PIL.Image.Image.entropy, is_method=True)
def entropy(self: PIL.Image.Image, mask: Optional[PIL.Image.Image] = None, extrema: Optional[list] = None) -> float:
    """
    Returns the entropy of the image, optionally using a mask and extrema.

    Equivalent to
    [`PIL.Image.Image.entropy()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.entropy)

    Args:
        mask: An optional mask image.
        extrema: An optional list of extrema.
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.getbands, is_method=True)
def getbands(self: PIL.Image.Image) -> tuple[str]:
    """
    Return a tuple containing the names of the image bands.

    Equivalent to
    [`PIL.Image.Image.getbands()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getbands)
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.getbbox, is_method=True)
def getbbox(self: PIL.Image.Image, *, alpha_only: bool = True) -> tuple[int, int, int, int]:
    """
    Return a bounding box for the non-zero regions of the image.

    Equivalent to [`PIL.Image.Image.getbbox()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getbbox)

    Args:
        alpha_only: If `True`, and the image has an alpha channel, trim transparent pixels. Otherwise,
            trim pixels when all channels are zero.
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.getcolors, is_method=True)
def getcolors(self: PIL.Image.Image, maxcolors: int = 256) -> tuple[tuple[int, int, int], int]:
    """
    Return a list of colors used in the image, up to a maximum of `maxcolors`.

    Equivalent to
    [`PIL.Image.Image.getcolors()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getcolors)

    Args:
        maxcolors: The maximum number of colors to return.
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.getextrema, is_method=True)
def getextrema(self: PIL.Image.Image) -> tuple[int, int]:
    """
    Return a 2-tuple containing the minimum and maximum pixel values of the image.

    Equivalent to
    [`PIL.Image.Image.getextrema()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getextrema)
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.getpalette, is_method=True)
def getpalette(self: PIL.Image.Image, mode: Optional[str] = None) -> tuple[int]:
    """
    Return the palette of the image, optionally converting it to a different mode.

    Equivalent to
    [`PIL.Image.Image.getpalette()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getpalette)

    Args:
        mode: The mode to convert the palette to.
    """
    pass


@pxt.udf(is_method=True)
def getpixel(self: PIL.Image.Image, xy: list) -> tuple[int]:
    """
    Return the pixel value at the given position. The position `xy` is a tuple containing the x and y coordinates.

    Equivalent to
    [`PIL.Image.Image.getpixel()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getpixel)

    Args:
        xy: The coordinates, given as (x, y).
    """
    # `xy` will be a list; `tuple(xy)` is necessary for pillow 9 compatibility
    return self.getpixel(tuple(xy))


@pxt.udf(substitute_fn=PIL.Image.Image.getprojection, is_method=True)
def getprojection(self: PIL.Image.Image) -> tuple[int]:
    """
    Return two sequences representing the horizontal and vertical projection of the image.

    Equivalent to
    [`PIL.Image.Image.getprojection()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.getprojection)
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.histogram, is_method=True)
def histogram(
    self: PIL.Image.Image, mask: Optional[PIL.Image.Image] = None, extrema: Optional[list] = None
) -> list[int]:
    """
    Return a histogram for the image.

    Equivalent to
    [`PIL.Image.Image.histogram()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.histogram)

    Args:
        mask: An optional mask image.
        extrema: An optional list of extrema.
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.quantize, is_method=True)
def quantize(
    self: PIL.Image.Image,
    colors: int = 256,
    method: Optional[int] = None,
    kmeans: int = 0,
    palette: Optional[int] = None,
    dither: int = PIL.Image.Dither.FLOYDSTEINBERG,
) -> PIL.Image.Image:
    """
     Convert the image to 'P' mode with the specified number of colors.

     Equivalent to
     [`PIL.Image.Image.quantize()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.quantize)

    Args:
        colors: The number of colors to quantize to.
        method: The quantization method. See the
            [Pillow documentation](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.quantize)
            for a list of supported methods.
        kmeans: The number of k-means clusters to use.
        palette: The palette to use.
        dither: The dithering method. See the
            [Pillow documentation](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.quantize)
            for a list of supported methods.
    """
    pass


@pxt.udf(substitute_fn=PIL.Image.Image.reduce, is_method=True)
def reduce(self: PIL.Image.Image, factor: int, box: Optional[tuple[int, int, int, int]] = None) -> PIL.Image.Image:
    """
    Reduce the image by the given factor.

    Equivalent to
    [`PIL.Image.Image.reduce()`](https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.reduce)

    Args:
        factor: The reduction factor.
        box: An optional 4-tuple of ints providing the source image region to be reduced. The values must be within
            (0, 0, width, height) rectangle. If omitted or None, the entire source is used.
    """
    pass


@pxt.udf(is_property=True)
def width(self: PIL.Image.Image) -> int:
    return self.width


@pxt.udf(is_property=True)
def height(self: PIL.Image.Image) -> int:
    return self.height


@pxt.udf(is_property=True)
def mode(self: PIL.Image.Image) -> str:
    return self.mode


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import abc
import builtins
import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Literal, Optional, Union, overload

from typing import _GenericAlias  # type: ignore[attr-defined]  # isort: skip
from keyword import iskeyword as is_python_keyword
from uuid import UUID

import pandas as pd
import sqlalchemy as sql

import pixeltable as pxt
from pixeltable import catalog, env, exceptions as excs, exprs, index, type_system as ts
from pixeltable.env import Env
from pixeltable.metadata import schema

from ..exprs import ColumnRef
from ..utils.description_helper import DescriptionHelper
from ..utils.filecache import FileCache
from .column import Column
from .globals import (
    _ROWID_COLUMN_NAME,
    IfExistsParam,
    IfNotExistsParam,
    MediaValidation,
    UpdateStatus,
    is_system_column_name,
    is_valid_identifier,
)
from .schema_object import SchemaObject
from .table_version_handle import TableVersionHandle
from .table_version_path import TableVersionPath

if TYPE_CHECKING:
    import torch.utils.data

    import pixeltable.plan
    from pixeltable.globals import TableDataSource

_logger = logging.getLogger('pixeltable')


class Table(SchemaObject):
    """
    A handle to a table, view, or snapshot. This class is the primary interface through which table operations
    (queries, insertions, updates, etc.) are performed in Pixeltable.
    """

    # Every user-invoked operation that runs an ExecNode tree (directly or indirectly) needs to call
    # FileCache.emit_eviction_warnings() at the end of the operation.

    _is_dropped: bool
    __tbl_version_path: TableVersionPath

    def __init__(self, id: UUID, dir_id: UUID, name: str, tbl_version_path: TableVersionPath):
        super().__init__(id, name, dir_id)
        self._is_dropped = False
        self.__tbl_version_path = tbl_version_path

    # @property
    # def _has_dependents(self) -> bool:
    #     """Returns True if this table has any dependent views, or snapshots."""
    #     return len(self._get_views(recursive=False)) > 0

    def _move(self, new_name: str, new_dir_id: UUID) -> None:
        self._check_is_dropped()
        super()._move(new_name, new_dir_id)
        conn = env.Env.get().conn
        stmt = sql.text(
            (
                f'UPDATE {schema.Table.__table__} '
                f'SET {schema.Table.dir_id.name} = :new_dir_id, '
                f"    {schema.Table.md.name} = jsonb_set({schema.Table.md.name}, '{{name}}', (:new_name)::jsonb) "
                f'WHERE {schema.Table.id.name} = :id'
            )
        )
        conn.execute(stmt, {'new_dir_id': new_dir_id, 'new_name': json.dumps(new_name), 'id': self._id})

    def get_metadata(self) -> dict[str, Any]:
        """
        Retrieves metadata associated with this table.

        Returns:
            A dictionary containing the metadata, in the following format:

                ```python
                {
                    'base': None,  # If this is a view or snapshot, will contain the name of its base table
                    'schema': {
                        'col1': StringType(),
                        'col2': IntType(),
                    },
                    'version': 22,
                    'schema_version': 1,
                    'comment': '',
                    'num_retained_versions': 10,
                    'is_view': False,
                    'is_snapshot': False,
                    'media_validation': 'on_write',
                }
                ```
        """
        self._check_is_dropped()
        with env.Env.get().begin_xact():
            md = super().get_metadata()
            md['base'] = self._base._path() if self._base is not None else None
            md['schema'] = self._schema
            md['version'] = self._version
            md['schema_version'] = self._tbl_version.get().schema_version
            md['comment'] = self._comment
            md['num_retained_versions'] = self._num_retained_versions
            md['media_validation'] = self._media_validation.name.lower()
            return md

    @property
    def _version(self) -> int:
        """Return the version of this table. Used by tests to ascertain version changes."""
        return self._tbl_version.get().version

    @property
    def _tbl_version(self) -> TableVersionHandle:
        """Return TableVersion for just this table."""
        return self._tbl_version_path.tbl_version

    @property
    def _tbl_version_path(self) -> TableVersionPath:
        self._check_is_dropped()
        return self.__tbl_version_path

    def __hash__(self) -> int:
        return hash(self._tbl_version.id)

    def _check_is_dropped(self) -> None:
        if self._is_dropped:
            raise excs.Error(f'{self._display_name()} {self._name} has been dropped')

    def __getattr__(self, name: str) -> 'pxt.exprs.ColumnRef':
        """Return a ColumnRef for the given name."""
        col = self._tbl_version_path.get_column(name)
        if col is None:
            raise AttributeError(f'Column {name!r} unknown')
        return ColumnRef(col)

    def __getitem__(self, name: str) -> 'pxt.exprs.ColumnRef':
        """Return a ColumnRef for the given name."""
        return getattr(self, name)

    def list_views(self, *, recursive: bool = True) -> list[str]:
        """
        Returns a list of all views and snapshots of this `Table`.

        Args:
            recursive: If `False`, returns only the immediate successor views of this `Table`. If `True`, returns
                all sub-views (including views of views, etc.)

        Returns:
            A list of view paths.
        """
        self._check_is_dropped()
        with env.Env.get().begin_xact():
            return [t._path() for t in self._get_views(recursive=recursive)]

    def _get_views(self, *, recursive: bool = True) -> list['Table']:
        cat = catalog.Catalog.get()
        view_ids = cat.get_view_ids(self._id)
        views = [cat.get_table_by_id(id) for id in view_ids]
        if recursive:
            views.extend([t for view in views for t in view._get_views(recursive=True)])
        return views

    def _df(self) -> 'pxt.dataframe.DataFrame':
        """Return a DataFrame for this table."""
        # local import: avoid circular imports
        from pixeltable.plan import FromClause

        return pxt.DataFrame(FromClause(tbls=[self._tbl_version_path]))

    def select(self, *items: Any, **named_items: Any) -> 'pxt.DataFrame':
        """Select columns or expressions from this table.

        See [`DataFrame.select`][pixeltable.DataFrame.select] for more details.
        """
        return self._df().select(*items, **named_items)

    def where(self, pred: 'exprs.Expr') -> 'pxt.DataFrame':
        """Filter rows from this table based on the expression.

        See [`DataFrame.where`][pixeltable.DataFrame.where] for more details.
        """
        return self._df().where(pred)

    def join(
        self,
        other: 'Table',
        *,
        on: Optional['exprs.Expr'] = None,
        how: 'pixeltable.plan.JoinType.LiteralType' = 'inner',
    ) -> 'pxt.DataFrame':
        """Join this table with another table."""
        return self._df().join(other, on=on, how=how)

    def order_by(self, *items: 'exprs.Expr', asc: bool = True) -> 'pxt.DataFrame':
        """Order the rows of this table based on the expression.

        See [`DataFrame.order_by`][pixeltable.DataFrame.order_by] for more details.
        """
        return self._df().order_by(*items, asc=asc)

    def group_by(self, *items: 'exprs.Expr') -> 'pxt.DataFrame':
        """Group the rows of this table based on the expression.

        See [`DataFrame.group_by`][pixeltable.DataFrame.group_by] for more details.
        """
        return self._df().group_by(*items)

    def limit(self, n: int) -> 'pxt.DataFrame':
        return self._df().limit(n)

    def collect(self) -> 'pxt.dataframe.DataFrameResultSet':
        """Return rows from this table."""
        return self._df().collect()

    def show(self, *args: Any, **kwargs: Any) -> 'pxt.dataframe.DataFrameResultSet':
        """Return rows from this table."""
        return self._df().show(*args, **kwargs)

    def head(self, *args: Any, **kwargs: Any) -> 'pxt.dataframe.DataFrameResultSet':
        """Return the first n rows inserted into this table."""
        return self._df().head(*args, **kwargs)

    def tail(self, *args: Any, **kwargs: Any) -> 'pxt.dataframe.DataFrameResultSet':
        """Return the last n rows inserted into this table."""
        return self._df().tail(*args, **kwargs)

    def count(self) -> int:
        """Return the number of rows in this table."""
        return self._df().count()

    @property
    def columns(self) -> list[str]:
        """Return the names of the columns in this table."""
        cols = self._tbl_version_path.columns()
        return [c.name for c in cols]

    @property
    def _schema(self) -> dict[str, ts.ColumnType]:
        """Return the schema (column names and column types) of this table."""
        return {c.name: c.col_type for c in self._tbl_version_path.columns()}

    @property
    def _base(self) -> Optional['Table']:
        """
        The base table of this `Table`. If this table is a view, returns the `Table`
        from which it was derived. Otherwise, returns `None`.
        """
        if self._tbl_version_path.base is None:
            return None
        base_id = self._tbl_version_path.base.tbl_version.id
        return catalog.Catalog.get().get_table_by_id(base_id)

    @property
    def _bases(self) -> list['Table']:
        """
        The ancestor list of bases of this table, starting with its immediate base.
        """
        bases = []
        base = self._base
        while base is not None:
            bases.append(base)
            base = base._base
        return bases

    @property
    def _comment(self) -> str:
        return self._tbl_version.get().comment

    @property
    def _num_retained_versions(self) -> int:
        return self._tbl_version.get().num_retained_versions

    @property
    def _media_validation(self) -> MediaValidation:
        return self._tbl_version.get().media_validation

    def __repr__(self) -> str:
        return self._descriptors().to_string()

    def _repr_html_(self) -> str:
        return self._descriptors().to_html()

    def _descriptors(self) -> DescriptionHelper:
        """
        Constructs a list of descriptors for this table that can be pretty-printed.
        """
        helper = DescriptionHelper()
        helper.append(self._title_descriptor())
        helper.append(self._col_descriptor())
        idxs = self._index_descriptor()
        if not idxs.empty:
            helper.append(idxs)
        stores = self._external_store_descriptor()
        if not stores.empty:
            helper.append(stores)
        if self._comment:
            helper.append(f'COMMENT: {self._comment}')
        return helper

    def _title_descriptor(self) -> str:
        title: str
        if self._base is None:
            title = f'Table\n{self._path()!r}'
        else:
            title = f'View\n{self._path()!r}'
            title += f'\n(of {self.__bases_to_desc()})'
        return title

    def _col_descriptor(self, columns: Optional[list[str]] = None) -> pd.DataFrame:
        return pd.DataFrame(
            {
                'Column Name': col.name,
                'Type': col.col_type._to_str(as_schema=True),
                'Computed With': col.value_expr.display_str(inline=False) if col.value_expr is not None else '',
            }
            for col in self._tbl_version_path.columns()
            if columns is None or col.name in columns
        )

    def __bases_to_desc(self) -> str:
        bases = self._bases
        assert len(bases) >= 1
        if len(bases) <= 2:
            return ', '.join(repr(b._path()) for b in bases)
        else:
            return f'{bases[0]._path()!r}, ..., {bases[-1]._path()!r}'

    def _index_descriptor(self, columns: Optional[list[str]] = None) -> pd.DataFrame:
        from pixeltable import index

        pd_rows = []
        for name, info in self._tbl_version.get().idxs_by_name.items():
            if isinstance(info.idx, index.EmbeddingIndex) and (columns is None or info.col.name in columns):
                display_embed = info.idx.string_embed if info.col.col_type.is_string_type() else info.idx.image_embed
                if info.idx.string_embed is not None and info.idx.image_embed is not None:
                    embed_str = f'{display_embed} (+1)'
                else:
                    embed_str = str(display_embed)
                row = {
                    'Index Name': name,
                    'Column': info.col.name,
                    'Metric': str(info.idx.metric.name.lower()),
                    'Embedding': embed_str,
                }
                pd_rows.append(row)
        return pd.DataFrame(pd_rows)

    def _external_store_descriptor(self) -> pd.DataFrame:
        pd_rows = []
        for name, store in self._tbl_version.get().external_stores.items():
            row = {'External Store': name, 'Type': type(store).__name__}
            pd_rows.append(row)
        return pd.DataFrame(pd_rows)

    def describe(self) -> None:
        """
        Print the table schema.
        """
        self._check_is_dropped()
        if getattr(builtins, '__IPYTHON__', False):
            from IPython.display import display

            display(self._repr_html_())
        else:
            print(repr(self))

    def _drop(self) -> None:
        self._check_is_dropped()
        self._tbl_version.get().drop()
        self._is_dropped = True

    # TODO Factor this out into a separate module.
    # The return type is unresolvable, but torch can't be imported since it's an optional dependency.
    def to_pytorch_dataset(self, image_format: str = 'pt') -> 'torch.utils.data.IterableDataset':
        """Return a PyTorch Dataset for this table.
        See DataFrame.to_pytorch_dataset()
        """
        return self._df().to_pytorch_dataset(image_format=image_format)

    def to_coco_dataset(self) -> Path:
        """Return the path to a COCO json file for this table.
        See DataFrame.to_coco_dataset()
        """
        return self._df().to_coco_dataset()

    def _column_has_dependents(self, col: Column) -> bool:
        """Returns True if the column has dependents, False otherwise."""
        assert col is not None
        assert col.name in self._schema
        if any(c.name is not None for c in col.dependent_cols):
            return True
        return any(
            col in store.get_local_columns()
            for view in (self, *self._get_views(recursive=True))
            for store in view._tbl_version.get().external_stores.values()
        )

    def _ignore_or_drop_existing_columns(self, new_col_names: list[str], if_exists: IfExistsParam) -> list[str]:
        """Check and handle existing columns in the new column specification based on the if_exists parameter.

        If `if_exists='ignore'`, returns a list of existing columns, if any, in `new_col_names`.
        """
        assert not self.get_metadata()['is_snapshot']
        existing_col_names = set(self._schema.keys())
        cols_to_ignore = []
        for new_col_name in new_col_names:
            if new_col_name in existing_col_names:
                if if_exists == IfExistsParam.ERROR:
                    raise excs.Error(f'Duplicate column name: {new_col_name!r}')
                elif if_exists == IfExistsParam.IGNORE:
                    cols_to_ignore.append(new_col_name)
                elif if_exists in (IfExistsParam.REPLACE, IfExistsParam.REPLACE_FORCE):
                    if new_col_name not in self._tbl_version.get().cols_by_name:
                        # for views, it is possible that the existing column
                        # is a base table column; in that case, we should not
                        # drop/replace that column. Continue to raise error.
                        raise excs.Error(f'Column {new_col_name!r} is a base table column. Cannot replace it.')
                    col = self._tbl_version.get().cols_by_name[new_col_name]
                    # cannot drop a column with dependents; so reject
                    # replace directive if column has dependents.
                    if self._column_has_dependents(col):
                        raise excs.Error(
                            f'Column {new_col_name!r} already exists and has dependents. '
                            f'Cannot {if_exists.name.lower()} it.'
                        )
                    self.drop_column(new_col_name)
                    assert new_col_name not in self._tbl_version.get().cols_by_name
        return cols_to_ignore

    def add_columns(
        self,
        schema: dict[str, Union[ts.ColumnType, builtins.type, _GenericAlias]],
        if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
    ) -> UpdateStatus:
        """
        Adds multiple columns to the table. The columns must be concrete (non-computed) columns; to add computed
        columns, use [`add_computed_column()`][pixeltable.catalog.Table.add_computed_column] instead.

        The format of the `schema` argument is identical to the format of the schema in a call to
        [`create_table()`][pixeltable.globals.create_table].

        Args:
            schema: A dictionary mapping column names to types.
            if_exists: Determines the behavior if a column already exists. Must be one of the following:

                - `'error'`: an exception will be raised.
                - `'ignore'`: do nothing and return.
                - `'replace' or 'replace_force'`: drop the existing column and add the new column, if it has no
                    dependents.

                Note that the `if_exists` parameter is applied to all columns in the schema.
                To apply different behaviors to different columns, please use
                [`add_column()`][pixeltable.Table.add_column] for each column.

        Returns:
            Information about the execution status of the operation.

        Raises:
            Error: If any column name is invalid, or already exists and `if_exists='error'`,
                or `if_exists='replace*'` but the column has dependents or is a basetable column.

        Examples:
            Add multiple columns to the table `my_table`:

            >>> tbl = pxt.get_table('my_table')
            ... schema = {
            ...     'new_col_1': pxt.Int,
            ...     'new_col_2': pxt.String,
            ... }
            ... tbl.add_columns(schema)
        """
        self._check_is_dropped()
        if self.get_metadata()['is_snapshot']:
            raise excs.Error('Cannot add column to a snapshot.')
        col_schema = {
            col_name: {'type': ts.ColumnType.normalize_type(spec, nullable_default=True, allow_builtin_types=False)}
            for col_name, spec in schema.items()
        }

        with Env.get().begin_xact():
            # handle existing columns based on if_exists parameter
            cols_to_ignore = self._ignore_or_drop_existing_columns(
                list(col_schema.keys()), IfExistsParam.validated(if_exists, 'if_exists')
            )
            # if all columns to be added already exist and user asked to ignore
            # existing columns, there's nothing to do.
            for cname in cols_to_ignore:
                assert cname in col_schema
                del col_schema[cname]
            if len(col_schema) == 0:
                return UpdateStatus()
            new_cols = self._create_columns(col_schema)
            for new_col in new_cols:
                self._verify_column(new_col)
            status = self._tbl_version.get().add_columns(new_cols, print_stats=False, on_error='abort')
            FileCache.get().emit_eviction_warnings()
            return status

    def add_column(
        self,
        *,
        if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
        **kwargs: Union[ts.ColumnType, builtins.type, _GenericAlias, exprs.Expr],
    ) -> UpdateStatus:
        """
        Adds an ordinary (non-computed) column to the table.

        Args:
            kwargs: Exactly one keyword argument of the form `col_name=col_type`.
            if_exists: Determines the behavior if the column already exists. Must be one of the following:

                - `'error'`: an exception will be raised.
                - `'ignore'`: do nothing and return.
                - `'replace' or 'replace_force'`: drop the existing column and add the new column, if it has
                    no dependents.

        Returns:
            Information about the execution status of the operation.

        Raises:
            Error: If the column name is invalid, or already exists and `if_exists='erorr'`,
                or `if_exists='replace*'` but the column has dependents or is a basetable column.

        Examples:
            Add an int column:

            >>> tbl.add_column(new_col=pxt.Int)

            Alternatively, this can also be expressed as:

            >>> tbl.add_columns({'new_col': pxt.Int})
        """
        self._check_is_dropped()
        # verify kwargs
        if self._tbl_version.get().is_snapshot:
            raise excs.Error('Cannot add column to a snapshot.')
        # verify kwargs and construct column schema dict
        if len(kwargs) != 1:
            raise excs.Error(
                f'add_column() requires exactly one keyword argument of the form "col_name=col_type"; '
                f'got {len(kwargs)} instead ({", ".join(kwargs.keys())})'
            )
        col_type = next(iter(kwargs.values()))
        if not isinstance(col_type, (ts.ColumnType, type, _GenericAlias)):
            raise excs.Error(
                'The argument to add_column() must be a type; did you intend to use add_computed_column() instead?'
            )
        return self.add_columns(kwargs, if_exists=if_exists)

    def add_computed_column(
        self,
        *,
        stored: Optional[bool] = None,
        print_stats: bool = False,
        on_error: Literal['abort', 'ignore'] = 'abort',
        if_exists: Literal['error', 'ignore', 'replace'] = 'error',
        **kwargs: exprs.Expr,
    ) -> UpdateStatus:
        """
        Adds a computed column to the table.

        Args:
            kwargs: Exactly one keyword argument of the form `col_name=expression`.
            stored: Whether the column is materialized and stored or computed on demand.
            print_stats: If `True`, print execution metrics during evaluation.
            on_error: Determines the behavior if an error occurs while evaluating the column expression for at least one
                row.

                - `'abort'`: an exception will be raised and the column will not be added.
                - `'ignore'`: execution will continue and the column will be added. Any rows
                    with errors will have a `None` value for the column, with information about the error stored in the
                    corresponding `tbl.col_name.errortype` and `tbl.col_name.errormsg` fields.
            if_exists: Determines the behavior if the column already exists. Must be one of the following:

                - `'error'`: an exception will be raised.
                - `'ignore'`: do nothing and return.
                - `'replace' or 'replace_force'`: drop the existing column and add the new column, iff it has
                    no dependents.

        Returns:
            Information about the execution status of the operation.

        Raises:
            Error: If the column name is invalid or already exists and `if_exists='error'`,
                or `if_exists='replace*'` but the column has dependents or is a basetable column.

        Examples:
            For a table with an image column `frame`, add an image column `rotated` that rotates the image by
            90 degrees:

            >>> tbl.add_computed_column(rotated=tbl.frame.rotate(90))

            Do the same, but now the column is unstored:

            >>> tbl.add_computed_column(rotated=tbl.frame.rotate(90), stored=False)
        """
        self._check_is_dropped()
        if self.get_metadata()['is_snapshot']:
            raise excs.Error('Cannot add column to a snapshot.')
        if len(kwargs) != 1:
            raise excs.Error(
                f'add_computed_column() requires exactly one keyword argument of the form '
                '"column-name=type|value-expression"; '
                f'got {len(kwargs)} arguments instead ({", ".join(list(kwargs.keys()))})'
            )
        col_name, spec = next(iter(kwargs.items()))
        if not is_valid_identifier(col_name):
            raise excs.Error(f'Invalid column name: {col_name!r}')

        col_schema: dict[str, Any] = {'value': spec}
        if stored is not None:
            col_schema['stored'] = stored

        # Raise an error if the column expression refers to a column error property
        if isinstance(spec, exprs.Expr):
            for e in spec.subexprs(expr_class=exprs.ColumnPropertyRef, traverse_matches=False):
                if e.is_error_prop():
                    raise excs.Error(
                        'Use of a reference to an error property of another column is not allowed in a computed '
                        f'column. The specified computation for this column contains this reference: `{e!r}`'
                    )

        with Env.get().begin_xact():
            # handle existing columns based on if_exists parameter
            cols_to_ignore = self._ignore_or_drop_existing_columns(
                [col_name], IfExistsParam.validated(if_exists, 'if_exists')
            )
            # if the column to add already exists and user asked to ignore
            # exiting column, there's nothing to do.
            if len(cols_to_ignore) != 0:
                assert cols_to_ignore[0] == col_name
                return UpdateStatus()

            new_col = self._create_columns({col_name: col_schema})[0]
            self._verify_column(new_col)
            status = self._tbl_version.get().add_columns([new_col], print_stats=print_stats, on_error=on_error)
            FileCache.get().emit_eviction_warnings()
            return status

    @classmethod
    def _validate_column_spec(cls, name: str, spec: dict[str, Any]) -> None:
        """Check integrity of user-supplied Column spec

        We unfortunately can't use something like jsonschema for validation, because this isn't strictly a JSON schema
        (on account of containing Python Callables or Exprs).
        """
        assert isinstance(spec, dict)
        valid_keys = {'type', 'value', 'stored', 'media_validation'}
        for k in spec:
            if k not in valid_keys:
                raise excs.Error(f'Column {name}: invalid key {k!r}')

        if 'type' not in spec and 'value' not in spec:
            raise excs.Error(f"Column {name}: 'type' or 'value' must be specified")

        if 'type' in spec and not isinstance(spec['type'], (ts.ColumnType, type, _GenericAlias)):
            raise excs.Error(f'Column {name}: "type" must be a type or ColumnType, got {spec["type"]}')

        if 'value' in spec:
            value_expr = exprs.Expr.from_object(spec['value'])
            if value_expr is None:
                raise excs.Error(f'Column {name}: value must be a Pixeltable expression.')
            if 'type' in spec:
                raise excs.Error(f"Column {name}: 'type' is redundant if 'value' is specified")

        if 'media_validation' in spec:
            _ = catalog.MediaValidation.validated(spec['media_validation'], f'Column {name}: media_validation')

        if 'stored' in spec and not isinstance(spec['stored'], bool):
            raise excs.Error(f'Column {name}: "stored" must be a bool, got {spec["stored"]}')

    @classmethod
    def _create_columns(cls, schema: dict[str, Any]) -> list[Column]:
        """Construct list of Columns, given schema"""
        columns: list[Column] = []
        for name, spec in schema.items():
            col_type: Optional[ts.ColumnType] = None
            value_expr: Optional[exprs.Expr] = None
            primary_key: Optional[bool] = None
            media_validation: Optional[catalog.MediaValidation] = None
            stored = True

            if isinstance(spec, (ts.ColumnType, type, _GenericAlias)):
                col_type = ts.ColumnType.normalize_type(spec, nullable_default=True, allow_builtin_types=False)
            elif isinstance(spec, exprs.Expr):
                # create copy so we can modify it
                value_expr = spec.copy()
                value_expr.bind_rel_paths()
            elif isinstance(spec, dict):
                cls._validate_column_spec(name, spec)
                if 'type' in spec:
                    col_type = ts.ColumnType.normalize_type(
                        spec['type'], nullable_default=True, allow_builtin_types=False
                    )
                value_expr = spec.get('value')
                if value_expr is not None and isinstance(value_expr, exprs.Expr):
                    # create copy so we can modify it
                    value_expr = value_expr.copy()
                    value_expr.bind_rel_paths()
                stored = spec.get('stored', True)
                primary_key = spec.get('primary_key')
                media_validation_str = spec.get('media_validation')
                media_validation = (
                    catalog.MediaValidation[media_validation_str.upper()] if media_validation_str is not None else None
                )
            else:
                raise excs.Error(f'Invalid value for column {name!r}')

            column = Column(
                name,
                col_type=col_type,
                computed_with=value_expr,
                stored=stored,
                is_pk=primary_key,
                media_validation=media_validation,
            )
            columns.append(column)
        return columns

    @classmethod
    def validate_column_name(cls, name: str) -> None:
        """Check that a name is usable as a pixeltalbe column name"""
        if is_system_column_name(name) or is_python_keyword(name):
            raise excs.Error(f'{name!r} is a reserved name in Pixeltable; please choose a different column name.')
        if not is_valid_identifier(name):
            raise excs.Error(f'Invalid column name: {name!r}')

    @classmethod
    def _verify_column(cls, col: Column) -> None:
        """Check integrity of user-supplied Column and supply defaults"""
        cls.validate_column_name(col.name)
        if col.stored is False and not col.is_computed:
            raise excs.Error(f'Column {col.name!r}: stored={col.stored} only applies to computed columns')
        if col.stored is False and col.has_window_fn_call():
            raise excs.Error(
                (
                    f'Column {col.name!r}: stored={col.stored} is not valid for image columns computed with a '
                    f'streaming function'
                )
            )

    @classmethod
    def _verify_schema(cls, schema: list[Column]) -> None:
        """Check integrity of user-supplied schema and set defaults"""
        column_names: set[str] = set()
        for col in schema:
            cls._verify_column(col)
            column_names.add(col.name)

    def drop_column(self, column: Union[str, ColumnRef], if_not_exists: Literal['error', 'ignore'] = 'error') -> None:
        """Drop a column from the table.

        Args:
            column: The name or reference of the column to drop.
            if_not_exists: Directive for handling a non-existent column. Must be one of the following:

                - `'error'`: raise an error if the column does not exist.
                - `'ignore'`: do nothing if the column does not exist.

        Raises:
            Error: If the column does not exist and `if_exists='error'`,
                or if it is referenced by a dependent computed column.

        Examples:
            Drop the column `col` from the table `my_table` by column name:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_column('col')

            Drop the column `col` from the table `my_table` by column reference:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_column(tbl.col)

            Drop the column `col` from the table `my_table` if it exists, otherwise do nothing:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_col(tbl.col, if_not_exists='ignore')
        """
        self._check_is_dropped()
        if self._tbl_version_path.is_snapshot():
            raise excs.Error('Cannot drop column from a snapshot.')
        col: Column = None
        if_not_exists_ = IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
        if isinstance(column, str):
            col = self._tbl_version_path.get_column(column, include_bases=False)
            if col is None:
                if if_not_exists_ == IfNotExistsParam.ERROR:
                    raise excs.Error(f'Column {column!r} unknown')
                assert if_not_exists_ == IfNotExistsParam.IGNORE
                return
            col = self._tbl_version.get().cols_by_name[column]
        else:
            exists = self._tbl_version_path.has_column(column.col, include_bases=False)
            if not exists:
                if if_not_exists_ == IfNotExistsParam.ERROR:
                    raise excs.Error(f'Unknown column: {column.col.qualified_name}')
                assert if_not_exists_ == IfNotExistsParam.IGNORE
                return
            col = column.col

        dependent_user_cols = [c for c in col.dependent_cols if c.name is not None]
        if len(dependent_user_cols) > 0:
            raise excs.Error(
                f'Cannot drop column `{col.name}` because the following columns depend on it:\n'
                f'{", ".join(c.name for c in dependent_user_cols)}'
            )

        with Env.get().begin_xact():
            # See if this column has a dependent store. We need to look through all stores in all
            # (transitive) views of this table.
            dependent_stores = [
                (view, store)
                for view in (self, *self._get_views(recursive=True))
                for store in view._tbl_version.get().external_stores.values()
                if col in store.get_local_columns()
            ]
            if len(dependent_stores) > 0:
                dependent_store_names = [
                    store.name if view._id == self._id else f'{store.name} (in view `{view._name}`)'
                    for view, store in dependent_stores
                ]
                raise excs.Error(
                    f'Cannot drop column `{col.name}` because the following external stores depend on it:\n'
                    f'{", ".join(dependent_store_names)}'
                )

            self._tbl_version.get().drop_column(col)

    def rename_column(self, old_name: str, new_name: str) -> None:
        """Rename a column.

        Args:
            old_name: The current name of the column.
            new_name: The new name of the column.

        Raises:
            Error: If the column does not exist, or if the new name is invalid or already exists.

        Examples:
            Rename the column `col1` to `col2` of the table `my_table`:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.rename_column('col1', 'col2')
        """
        with Env.get().begin_xact():
            self._tbl_version.get().rename_column(old_name, new_name)

    def _list_index_info_for_test(self) -> list[dict[str, Any]]:
        """
        Returns list of all the indexes on this table. Used for testing.

        Returns:
            A list of index information, each containing the index's
            id, name, and the name of the column it indexes.
        """
        assert not self._is_dropped
        index_info = []
        for idx_name, idx in self._tbl_version.get().idxs_by_name.items():
            index_info.append({'_id': idx.id, '_name': idx_name, '_column': idx.col.name})
        return index_info

    def add_embedding_index(
        self,
        column: Union[str, ColumnRef],
        *,
        idx_name: Optional[str] = None,
        embedding: Optional[pxt.Function] = None,
        string_embed: Optional[pxt.Function] = None,
        image_embed: Optional[pxt.Function] = None,
        metric: str = 'cosine',
        if_exists: Literal['error', 'ignore', 'replace', 'replace_force'] = 'error',
    ) -> None:
        """
        Add an embedding index to the table. Once the index is created, it will be automatically kept up-to-date as new
        rows are inserted into the table.

        To add an embedding index, one must specify, at minimum, the column to be indexed and an embedding UDF.
        Only `String` and `Image` columns are currently supported. Here's an example that uses a
        [CLIP embedding][pixeltable.functions.huggingface.clip] to index an image column:

        >>> from pixeltable.functions.huggingface import clip
        ... embedding_fn = clip.using(model_id='openai/clip-vit-base-patch32')
        ... tbl.add_embedding_index(tbl.img, embedding=embedding_fn)

        Once the index is created, similiarity lookups can be performed using the `similarity` pseudo-function.

        >>> reference_img = PIL.Image.open('my_image.jpg')
        ... sim = tbl.img.similarity(reference_img)
        ... tbl.select(tbl.img, sim).order_by(sim, asc=False).limit(5)

        If the embedding UDF is a multimodal embedding (supporting more than one data type), then lookups may be
        performed using any of its supported types. In our example, CLIP supports both text and images, so we can
        also search for images using a text description:

        >>> sim = tbl.img.similarity('a picture of a train')
        ... tbl.select(tbl.img, sim).order_by(sim, asc=False).limit(5)

        Args:
            column: The name of, or reference to, the column to be indexed; must be a `String` or `Image` column.
            idx_name: An optional name for the index. If not specified, a name such as `'idx0'` will be generated
                automatically. If specified, the name must be unique for this table and a valid pixeltable column name.
            embedding: The UDF to use for the embedding. Must be a UDF that accepts a single argument of type `String`
                or `Image` (as appropriate for the column being indexed) and returns a fixed-size 1-dimensional
                array of floats.
            string_embed: An optional UDF to use for the string embedding component of this index.
                Can be used in conjunction with `image_embed` to construct multimodal embeddings manually, by
                specifying different embedding functions for different data types.
            image_embed: An optional UDF to use for the image embedding component of this index.
                Can be used in conjunction with `string_embed` to construct multimodal embeddings manually, by
                specifying different embedding functions for different data types.
            metric: Distance metric to use for the index; one of `'cosine'`, `'ip'`, or `'l2'`.
                The default is `'cosine'`.
            if_exists: Directive for handling an existing index with the same name. Must be one of the following:

                - `'error'`: raise an error if an index with the same name already exists.
                - `'ignore'`: do nothing if an index with the same name already exists.
                - `'replace'` or `'replace_force'`: replace the existing index with the new one.

        Raises:
            Error: If an index with the specified name already exists for the table and `if_exists='error'`, or if
                the specified column does not exist.

        Examples:
            Add an index to the `img` column of the table `my_table`:

            >>> from pixeltable.functions.huggingface import clip
            ... tbl = pxt.get_table('my_table')
            ... embedding_fn = clip.using(model_id='openai/clip-vit-base-patch32')
            ... tbl.add_embedding_index(tbl.img, embedding=embedding_fn)

            Alternatively, the `img` column may be specified by name:

            >>> tbl.add_embedding_index('img', embedding=embedding_fn)

            Add a second index to the `img` column, using the inner product as the distance metric,
            and with a specific name:

            >>> tbl.add_embedding_index(
            ...     tbl.img,
            ...     idx_name='ip_idx',
            ...     embedding=embedding_fn,
            ...     metric='ip'
            ... )

            Add an index using separately specified string and image embeddings:

            >>> tbl.add_embedding_index(
            ...     tbl.img,
            ...     string_embed=string_embedding_fn,
            ...     image_embed=image_embedding_fn
            ... )
        """
        if self._tbl_version_path.is_snapshot():
            raise excs.Error('Cannot add an index to a snapshot')
        col = self._resolve_column_parameter(column)

        with Env.get().begin_xact():
            if idx_name is not None and idx_name in self._tbl_version.get().idxs_by_name:
                if_exists_ = IfExistsParam.validated(if_exists, 'if_exists')
                # An index with the same name already exists.
                # Handle it according to if_exists.
                if if_exists_ == IfExistsParam.ERROR:
                    raise excs.Error(f'Duplicate index name: {idx_name}')
                if not isinstance(self._tbl_version.get().idxs_by_name[idx_name].idx, index.EmbeddingIndex):
                    raise excs.Error(
                        f'Index `{idx_name}` is not an embedding index. Cannot {if_exists_.name.lower()} it.'
                    )
                if if_exists_ == IfExistsParam.IGNORE:
                    return
                assert if_exists_ in (IfExistsParam.REPLACE, IfExistsParam.REPLACE_FORCE)
                self.drop_index(idx_name=idx_name)
                assert idx_name not in self._tbl_version.get().idxs_by_name
            from pixeltable.index import EmbeddingIndex

            # idx_name must be a valid pixeltable column name
            if idx_name is not None:
                Table.validate_column_name(idx_name)

            # create the EmbeddingIndex instance to verify args
            idx = EmbeddingIndex(
                col, metric=metric, embed=embedding, string_embed=string_embed, image_embed=image_embed
            )
            _ = self._tbl_version.get().add_index(col, idx_name=idx_name, idx=idx)
            # TODO: how to deal with exceptions here? drop the index and raise?
            FileCache.get().emit_eviction_warnings()

    def drop_embedding_index(
        self,
        *,
        column: Union[str, ColumnRef, None] = None,
        idx_name: Optional[str] = None,
        if_not_exists: Literal['error', 'ignore'] = 'error',
    ) -> None:
        """
        Drop an embedding index from the table. Either a column name or an index name (but not both) must be
        specified. If a column name or reference is specified, it must be a column containing exactly one
        embedding index; otherwise the specific index name must be provided instead.

        Args:
            column: The name of, or reference to, the column from which to drop the index.
                    The column must have only one embedding index.
            idx_name: The name of the index to drop.
            if_not_exists: Directive for handling a non-existent index. Must be one of the following:

                - `'error'`: raise an error if the index does not exist.
                - `'ignore'`: do nothing if the index does not exist.

                Note that `if_not_exists` parameter is only applicable when an `idx_name` is specified
                and it does not exist, or when `column` is specified and it has no index.
                `if_not_exists` does not apply to non-exisitng column.

        Raises:
            Error: If `column` is specified, but the column does not exist, or it contains no embedding
                indices and `if_not_exists='error'`, or the column has multiple embedding indices.
            Error: If `idx_name` is specified, but the index is not an embedding index, or
                the index does not exist and `if_not_exists='error'`.

        Examples:
            Drop the embedding index on the `img` column of the table `my_table` by column name:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_embedding_index(column='img')

            Drop the embedding index on the `img` column of the table `my_table` by column reference:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_embedding_index(column=tbl.img)

            Drop the embedding index `idx1` of the table `my_table` by index name:
            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_embedding_index(idx_name='idx1')

            Drop the embedding index `idx1` of the table `my_table` by index name, if it exists, otherwise do nothing:
            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_embedding_index(idx_name='idx1', if_not_exists='ignore')
        """
        if (column is None) == (idx_name is None):
            raise excs.Error("Exactly one of 'column' or 'idx_name' must be provided")

        col: Column = None
        if idx_name is None:
            col = self._resolve_column_parameter(column)
            assert col is not None

        with Env.get().begin_xact():
            self._drop_index(col=col, idx_name=idx_name, _idx_class=index.EmbeddingIndex, if_not_exists=if_not_exists)

    def _resolve_column_parameter(self, column: Union[str, ColumnRef]) -> Column:
        """Resolve a column parameter to a Column object"""
        col: Column = None
        if isinstance(column, str):
            col = self._tbl_version_path.get_column(column, include_bases=True)
            if col is None:
                raise excs.Error(f'Column {column!r} unknown')
        elif isinstance(column, ColumnRef):
            exists = self._tbl_version_path.has_column(column.col, include_bases=True)
            if not exists:
                raise excs.Error(f'Unknown column: {column.col.qualified_name}')
            col = column.col
        else:
            raise excs.Error(f'Invalid column parameter type: {type(column)}')
        return col

    def drop_index(
        self,
        *,
        column: Union[str, ColumnRef, None] = None,
        idx_name: Optional[str] = None,
        if_not_exists: Literal['error', 'ignore'] = 'error',
    ) -> None:
        """
        Drop an index from the table. Either a column name or an index name (but not both) must be
        specified. If a column name or reference is specified, it must be a column containing exactly one index;
        otherwise the specific index name must be provided instead.

        Args:
            column: The name of, or reference to, the column from which to drop the index.
                    The column must have only one embedding index.
            idx_name: The name of the index to drop.
            if_not_exists: Directive for handling a non-existent index. Must be one of the following:

                - `'error'`: raise an error if the index does not exist.
                - `'ignore'`: do nothing if the index does not exist.

                Note that `if_not_exists` parameter is only applicable when an `idx_name` is specified
                and it does not exist, or when `column` is specified and it has no index.
                `if_not_exists` does not apply to non-exisitng column.

        Raises:
            Error: If `column` is specified, but the column does not exist, or it contains no
                indices or multiple indices.
            Error: If `idx_name` is specified, but the index does not exist.

        Examples:
            Drop the index on the `img` column of the table `my_table` by column name:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_index(column_name='img')

            Drop the index on the `img` column of the table `my_table` by column reference:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_index(tbl.img)

            Drop the index `idx1` of the table `my_table` by index name:
            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_index(idx_name='idx1')

            Drop the index `idx1` of the table `my_table` by index name, if it exists, otherwise do nothing:
            >>> tbl = pxt.get_table('my_table')
            ... tbl.drop_index(idx_name='idx1', if_not_exists='ignore')

        """
        if (column is None) == (idx_name is None):
            raise excs.Error("Exactly one of 'column' or 'idx_name' must be provided")

        col: Column = None
        if idx_name is None:
            col = self._resolve_column_parameter(column)
            assert col is not None

        with Env.get().begin_xact():
            self._drop_index(col=col, idx_name=idx_name, if_not_exists=if_not_exists)

    def _drop_index(
        self,
        *,
        col: Optional[Column] = None,
        idx_name: Optional[str] = None,
        _idx_class: Optional[type[index.IndexBase]] = None,
        if_not_exists: Literal['error', 'ignore'] = 'error',
    ) -> None:
        if self._tbl_version_path.is_snapshot():
            raise excs.Error('Cannot drop an index from a snapshot')
        assert (col is None) != (idx_name is None)

        if idx_name is not None:
            if_not_exists_ = IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
            if idx_name not in self._tbl_version.get().idxs_by_name:
                if if_not_exists_ == IfNotExistsParam.ERROR:
                    raise excs.Error(f'Index {idx_name!r} does not exist')
                assert if_not_exists_ == IfNotExistsParam.IGNORE
                return
            idx_info = self._tbl_version.get().idxs_by_name[idx_name]
        else:
            if col.tbl.id != self._tbl_version.id:
                raise excs.Error(
                    f'Column {col.name!r}: cannot drop index from column that belongs to base ({col.tbl.get().name}!r)'
                )
            idx_info_list = [info for info in self._tbl_version.get().idxs_by_name.values() if info.col.id == col.id]
            if _idx_class is not None:
                idx_info_list = [info for info in idx_info_list if isinstance(info.idx, _idx_class)]
            if len(idx_info_list) == 0:
                if_not_exists_ = IfNotExistsParam.validated(if_not_exists, 'if_not_exists')
                if if_not_exists_ == IfNotExistsParam.ERROR:
                    raise excs.Error(f'Column {col.name!r} does not have an index')
                assert if_not_exists_ == IfNotExistsParam.IGNORE
                return
            if len(idx_info_list) > 1:
                raise excs.Error(f"Column {col.name!r} has multiple indices; specify 'idx_name' instead")
            idx_info = idx_info_list[0]

        # Find out if anything depends on this index
        dependent_user_cols = [c for c in idx_info.val_col.dependent_cols if c.name is not None]
        if len(dependent_user_cols) > 0:
            raise excs.Error(
                f'Cannot drop index because the following columns depend on it:\n'
                f'{", ".join(c.name for c in dependent_user_cols)}'
            )
        self._tbl_version.get().drop_index(idx_info.id)

    @overload
    def insert(
        self,
        source: TableDataSource,
        /,
        *,
        source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
        schema_overrides: Optional[dict[str, ts.ColumnType]] = None,
        on_error: Literal['abort', 'ignore'] = 'abort',
        print_stats: bool = False,
        **kwargs: Any,
    ) -> UpdateStatus: ...

    @overload
    def insert(
        self, /, *, on_error: Literal['abort', 'ignore'] = 'abort', print_stats: bool = False, **kwargs: Any
    ) -> UpdateStatus: ...

    @abc.abstractmethod
    def insert(
        self,
        source: Optional[TableDataSource] = None,
        /,
        *,
        source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
        schema_overrides: Optional[dict[str, ts.ColumnType]] = None,
        on_error: Literal['abort', 'ignore'] = 'abort',
        print_stats: bool = False,
        **kwargs: Any,
    ) -> UpdateStatus:
        """Inserts rows into this table. There are two mutually exclusive call patterns:

        To insert multiple rows at a time:

        ```python
        insert(
            source: TableSourceDataType,
            /,
            *,
            on_error: Literal['abort', 'ignore'] = 'abort',
            print_stats: bool = False,
            **kwargs: Any,
        )```

        To insert just a single row, you can use the more concise syntax:

        ```python
        insert(
            *,
            on_error: Literal['abort', 'ignore'] = 'abort',
            print_stats: bool = False,
            **kwargs: Any
        )```

        Args:
            source: A data source from which data can be imported.
            kwargs: (if inserting a single row) Keyword-argument pairs representing column names and values.
                (if inserting multiple rows) Additional keyword arguments are passed to the data source.
            source_format: A hint about the format of the source data
            schema_overrides: If specified, then columns in `schema_overrides` will be given the specified types
            on_error: Determines the behavior if an error occurs while evaluating a computed column or detecting an
                invalid media file (such as a corrupt image) for one of the inserted rows.

                - If `on_error='abort'`, then an exception will be raised and the rows will not be inserted.
                - If `on_error='ignore'`, then execution will continue and the rows will be inserted. Any cells
                    with errors will have a `None` value for that cell, with information about the error stored in the
                    corresponding `tbl.col_name.errortype` and `tbl.col_name.errormsg` fields.
            print_stats: If `True`, print statistics about the cost of computed columns.

        Returns:
            An [`UpdateStatus`][pixeltable.UpdateStatus] object containing information about the update.

        Raises:
            Error: If one of the following conditions occurs:

                - The table is a view or snapshot.
                - The table has been dropped.
                - One of the rows being inserted does not conform to the table schema.
                - An error occurs during processing of computed columns, and `on_error='ignore'`.
                - An error occurs while importing data from a source, and `on_error='abort'`.

        Examples:
            Insert two rows into the table `my_table` with three int columns ``a``, ``b``, and ``c``.
            Column ``c`` is nullable:

            >>> tbl = pxt.get_table('my_table')
            ... tbl.insert([{'a': 1, 'b': 1, 'c': 1}, {'a': 2, 'b': 2}])

            Insert a single row using the alternative syntax:

            >>> tbl.insert(a=3, b=3, c=3)

            Insert rows from a CSV file:

            >>> tbl.insert(source='path/to/file.csv')
        """
        raise NotImplementedError

    def update(
        self, value_spec: dict[str, Any], where: Optional['pxt.exprs.Expr'] = None, cascade: bool = True
    ) -> UpdateStatus:
        """Update rows in this table.

        Args:
            value_spec: a dictionary mapping column names to literal values or Pixeltable expressions.
            where: a predicate to filter rows to update.
            cascade: if True, also update all computed columns that transitively depend on the updated columns.

        Examples:
            Set column `int_col` to 1 for all rows:

            >>> tbl.update({'int_col': 1})

            Set column `int_col` to 1 for all rows where `int_col` is 0:

            >>> tbl.update({'int_col': 1}, where=tbl.int_col == 0)

            Set `int_col` to the value of `other_int_col` + 1:

            >>> tbl.update({'int_col': tbl.other_int_col + 1})

            Increment `int_col` by 1 for all rows where `int_col` is 0:

            >>> tbl.update({'int_col': tbl.int_col + 1}, where=tbl.int_col == 0)
        """
        with Env.get().begin_xact():
            status = self._tbl_version.get().update(value_spec, where, cascade)
            FileCache.get().emit_eviction_warnings()
            return status

    def batch_update(
        self,
        rows: Iterable[dict[str, Any]],
        cascade: bool = True,
        if_not_exists: Literal['error', 'ignore', 'insert'] = 'error',
    ) -> UpdateStatus:
        """Update rows in this table.

        Args:
            rows: an Iterable of dictionaries containing values for the updated columns plus values for the primary key
                  columns.
            cascade: if True, also update all computed columns that transitively depend on the updated columns.
            if_not_exists: Specifies the behavior if a row to update does not exist:

                - `'error'`: Raise an error.
                - `'ignore'`: Skip the row silently.
                - `'insert'`: Insert the row.

        Examples:
            Update the `name` and `age` columns for the rows with ids 1 and 2 (assuming `id` is the primary key).
            If either row does not exist, this raises an error:

            >>> tbl.update([{'id': 1, 'name': 'Alice', 'age': 30}, {'id': 2, 'name': 'Bob', 'age': 40}])

            Update the `name` and `age` columns for the row with `id` 1 (assuming `id` is the primary key) and insert
            the row with new `id` 3 (assuming this key does not exist):

            >>> tbl.update(
                [{'id': 1, 'name': 'Alice', 'age': 30}, {'id': 3, 'name': 'Bob', 'age': 40}],
                if_not_exists='insert')
        """
        if self._tbl_version_path.is_snapshot():
            raise excs.Error('Cannot update a snapshot')
        rows = list(rows)

        row_updates: list[dict[Column, exprs.Expr]] = []
        pk_col_names = {c.name for c in self._tbl_version.get().primary_key_columns()}

        # pseudo-column _rowid: contains the rowid of the row to update and can be used instead of the primary key
        has_rowid = _ROWID_COLUMN_NAME in rows[0]
        rowids: list[tuple[int, ...]] = []
        if len(pk_col_names) == 0 and not has_rowid:
            raise excs.Error('Table must have primary key for batch update')

        for row_spec in rows:
            col_vals = self._tbl_version.get()._validate_update_spec(
                row_spec, allow_pk=not has_rowid, allow_exprs=False, allow_media=False
            )
            if has_rowid:
                # we expect the _rowid column to be present for each row
                assert _ROWID_COLUMN_NAME in row_spec
                rowids.append(row_spec[_ROWID_COLUMN_NAME])
            else:
                col_names = {col.name for col in col_vals}
                if any(pk_col_name not in col_names for pk_col_name in pk_col_names):
                    missing_cols = pk_col_names - {col.name for col in col_vals}
                    raise excs.Error(f'Primary key columns ({", ".join(missing_cols)}) missing in {row_spec}')
            row_updates.append(col_vals)

        with Env.get().begin_xact():
            status = self._tbl_version.get().batch_update(
                row_updates,
                rowids,
                error_if_not_exists=if_not_exists == 'error',
                insert_if_not_exists=if_not_exists == 'insert',
                cascade=cascade,
            )
            FileCache.get().emit_eviction_warnings()
            return status

    def delete(self, where: Optional['pxt.exprs.Expr'] = None) -> UpdateStatus:
        """Delete rows in this table.

        Args:
            where: a predicate to filter rows to delete.

        Examples:
            Delete all rows in a table:

            >>> tbl.delete()

            Delete all rows in a table where column `a` is greater than 5:

            >>> tbl.delete(tbl.a > 5)
        """
        raise NotImplementedError

    def revert(self) -> None:
        """Reverts the table to the previous version.

        .. warning::
            This operation is irreversible.
        """
        if self._tbl_version_path.is_snapshot():
            raise excs.Error('Cannot revert a snapshot')
        with Env.get().begin_xact():
            self._tbl_version.get().revert()

    @property
    def external_stores(self) -> list[str]:
        return list(self._tbl_version.get().external_stores.keys())

    def _link_external_store(self, store: 'pxt.io.ExternalStore') -> None:
        """
        Links the specified `ExternalStore` to this table.
        """
        if self._tbl_version.get().is_snapshot:
            raise excs.Error(f'Table `{self._name}` is a snapshot, so it cannot be linked to an external store.')
        if store.name in self.external_stores:
            raise excs.Error(f'Table `{self._name}` already has an external store with that name: {store.name}')
        _logger.info(f'Linking external store `{store.name}` to table `{self._name}`')
        with Env.get().begin_xact():
            self._tbl_version.get().link_external_store(store)
            env.Env.get().console_logger.info(f'Linked external store `{store.name}` to table `{self._name}`.')

    def unlink_external_stores(
        self,
        stores: Optional[str | list[str]] = None,
        *,
        delete_external_data: bool = False,
        ignore_errors: bool = False,
    ) -> None:
        """
        Unlinks this table's external stores.

        Args:
            stores: If specified, will unlink only the specified named store or list of stores. If not specified,
                will unlink all of this table's external stores.
            ignore_errors (bool): If `True`, no exception will be thrown if a specified store is not linked
                to this table.
            delete_external_data (bool): If `True`, then the external data store will also be deleted. WARNING: This
                is a destructive operation that will delete data outside Pixeltable, and cannot be undone.
        """
        self._check_is_dropped()
        all_stores = self.external_stores

        if stores is None:
            stores = all_stores
        elif isinstance(stores, str):
            stores = [stores]

        # Validation
        if not ignore_errors:
            for store in stores:
                if store not in all_stores:
                    raise excs.Error(f'Table `{self._name}` has no external store with that name: {store}')

        with Env.get().begin_xact():
            for store in stores:
                self._tbl_version.get().unlink_external_store(store, delete_external_data=delete_external_data)
                env.Env.get().console_logger.info(f'Unlinked external store from table `{self._name}`: {store}')

    def sync(
        self, stores: Optional[str | list[str]] = None, *, export_data: bool = True, import_data: bool = True
    ) -> 'pxt.io.SyncStatus':
        """
        Synchronizes this table with its linked external stores.

        Args:
            stores: If specified, will synchronize only the specified named store or list of stores. If not specified,
                will synchronize all of this table's external stores.
            export_data: If `True`, data from this table will be exported to the external stores during synchronization.
            import_data: If `True`, data from the external stores will be imported to this table during synchronization.
        """
        self._check_is_dropped()
        all_stores = self.external_stores

        if stores is None:
            stores = all_stores
        elif isinstance(stores, str):
            stores = [stores]

        for store in stores:
            if store not in all_stores:
                raise excs.Error(f'Table `{self._name}` has no external store with that name: {store}')

        sync_status = pxt.io.SyncStatus.empty()
        with Env.get().begin_xact():
            for store in stores:
                store_obj = self._tbl_version.get().external_stores[store]
                store_sync_status = store_obj.sync(self, export_data=export_data, import_data=import_data)
                sync_status = sync_status.combine(store_sync_status)

        return sync_status

    def __dir__(self) -> list[str]:
        return list(super().__dir__()) + list(self._schema.keys())

    def _ipython_key_completions_(self) -> list[str]:
        return list(self._schema.keys())



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `JsonType`.

Example:
```python
import pixeltable as pxt

t = pxt.get_table(...)
t.select(pxt.functions.json.make_list()).collect()
```
"""

from typing import Any

import pixeltable as pxt
from pixeltable.utils.code import local_public_names


@pxt.uda
class make_list(pxt.Aggregator):
    """
    Collects arguments into a list.
    """

    def __init__(self) -> None:
        self.output: list[Any] = []

    def update(self, obj: pxt.Json) -> None:
        if obj is None:
            return
        self.output.append(obj)

    def value(self) -> list[Any]:
        return self.output


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



# ruff: noqa: F401

from pixeltable.utils.code import local_public_names

from . import (
    anthropic,
    audio,
    deepseek,
    fireworks,
    gemini,
    huggingface,
    image,
    json,
    llama_cpp,
    math,
    mistralai,
    ollama,
    openai,
    replicate,
    string,
    timestamp,
    together,
    video,
    vision,
    whisper,
)
from .globals import count, map, max, mean, min, sum

__all__ = local_public_names(__name__, exclude=['globals']) + local_public_names(globals.__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import abc
import inspect
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Optional, Sequence, overload

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

from .function import Function
from .globals import validate_symbol_path
from .signature import Parameter, Signature

if TYPE_CHECKING:
    from pixeltable import exprs


class Aggregator(abc.ABC):
    @abc.abstractmethod
    def update(self, *args: Any, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def value(self) -> Any: ...


class AggregateFunction(Function):
    """Function interface for an aggregation operation.

    requires_order_by: if True, the first parameter to an aggregate function defines the order in which the function
    sees rows in update()
    allows_std_agg: if True, the aggregate function can be used as a standard aggregate function w/o a window
    allows_window: if True, the aggregate function can be used with a window
    """

    ORDER_BY_PARAM: ClassVar[str] = 'order_by'
    GROUP_BY_PARAM: ClassVar[str] = 'group_by'
    RESERVED_PARAMS: ClassVar[set[str]] = {ORDER_BY_PARAM, GROUP_BY_PARAM}

    agg_classes: list[type[Aggregator]]  # classes for each signature, in signature order
    init_param_names: list[list[str]]  # names of the __init__ parameters for each signature

    def __init__(
        self,
        agg_class: type[Aggregator],
        type_substitutions: Optional[Sequence[dict]],
        self_path: str,
        requires_order_by: bool,
        allows_std_agg: bool,
        allows_window: bool,
    ) -> None:
        if type_substitutions is None:
            type_substitutions = [None]  # single signature with no substitutions
            self.agg_classes = [agg_class]
        else:
            self.agg_classes = [agg_class] * len(type_substitutions)
        self.init_param_names = []
        self.requires_order_by = requires_order_by
        self.allows_std_agg = allows_std_agg
        self.allows_window = allows_window
        self.__doc__ = agg_class.__doc__

        signatures: list[Signature] = []

        # If no type_substitutions were provided, construct a single signature for the class.
        # Otherwise, construct one signature for each type substitution instance.
        for subst in type_substitutions:
            signature, init_param_names = self.__cls_to_signature(agg_class, subst)
            signatures.append(signature)
            self.init_param_names.append(init_param_names)

        super().__init__(signatures, self_path=self_path)

    def _update_as_overload_resolution(self, signature_idx: int) -> None:
        self.agg_classes = [self.agg_classes[signature_idx]]
        self.init_param_names = [self.init_param_names[signature_idx]]

    def __cls_to_signature(
        self, cls: type[Aggregator], type_substitutions: Optional[dict] = None
    ) -> tuple[Signature, list[str]]:
        """Inspects the Aggregator class to infer the corresponding function signature. Returns the
        inferred signature along with the list of init_param_names (for downstream error handling).
        """
        from pixeltable import exprs

        # infer type parameters; set return_type=InvalidType() because it has no meaning here
        init_sig = Signature.create(
            py_fn=cls.__init__, return_type=ts.InvalidType(), is_cls_method=True, type_substitutions=type_substitutions
        )
        update_sig = Signature.create(
            py_fn=cls.update, return_type=ts.InvalidType(), is_cls_method=True, type_substitutions=type_substitutions
        )
        value_sig = Signature.create(py_fn=cls.value, is_cls_method=True, type_substitutions=type_substitutions)

        init_types = [p.col_type for p in init_sig.parameters.values()]
        update_types = [p.col_type for p in update_sig.parameters.values()]
        value_type = value_sig.return_type
        assert value_type is not None

        if len(update_types) == 0:
            raise excs.Error('update() must have at least one parameter')

        # our signature is the signature of 'update', but without self,
        # plus the parameters of 'init' as keyword-only parameters
        py_update_params = list(inspect.signature(cls.update).parameters.values())[1:]  # leave out self
        assert len(py_update_params) == len(update_types)
        update_params = [
            Parameter(
                p.name,
                col_type=update_types[i],
                kind=p.kind,
                default=exprs.Expr.from_object(p.default),  # type: ignore[arg-type]
            )
            for i, p in enumerate(py_update_params)
        ]
        # starting at 1: leave out self
        py_init_params = list(inspect.signature(cls.__init__).parameters.values())[1:]
        assert len(py_init_params) == len(init_types)
        init_params = [
            Parameter(
                p.name,
                col_type=init_types[i],
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=exprs.Expr.from_object(p.default),  # type: ignore[arg-type]
            )
            for i, p in enumerate(py_init_params)
        ]
        duplicate_params = {p.name for p in init_params} & {p.name for p in update_params}
        if len(duplicate_params) > 0:
            raise excs.Error(
                f'__init__() and update() cannot have parameters with the same name: {", ".join(duplicate_params)}'
            )
        params = update_params + init_params  # init_params are keyword-only and come last
        init_param_names = [p.name for p in init_params]

        return Signature(value_type, params), init_param_names

    @property
    def agg_class(self) -> type[Aggregator]:
        assert not self.is_polymorphic
        return self.agg_classes[0]

    @property
    def is_async(self) -> bool:
        return False

    def exec(self, args: Sequence[Any], kwargs: dict[str, Any]) -> Any:
        raise NotImplementedError

    def overload(self, cls: type[Aggregator]) -> AggregateFunction:
        if not isinstance(cls, type) or not issubclass(cls, Aggregator):
            raise excs.Error(f'Invalid argument to @overload decorator: {cls}')
        if self._has_resolved_fns:
            raise excs.Error('New `overload` not allowed after the UDF has already been called')
        if self._conditional_return_type is not None:
            raise excs.Error('New `overload` not allowed after a conditional return type has been specified')
        sig, init_param_names = self.__cls_to_signature(cls)
        self.signatures.append(sig)
        self.agg_classes.append(cls)
        self.init_param_names.append(init_param_names)
        return self

    def _docstring(self) -> Optional[str]:
        return inspect.getdoc(self.agg_classes[0])

    def help_str(self) -> str:
        res = super().help_str()
        # We need to reference agg_classes[0] rather than agg_class here, because we want this to work even if the
        # aggregator is polymorphic (in which case we use the docstring of the originally decorated UDA).
        res += '\n\n' + inspect.getdoc(self.agg_classes[0].update)
        return res

    def __call__(self, *args: Any, **kwargs: Any) -> 'exprs.FunctionCall':
        from pixeltable import exprs

        # perform semantic analysis of special parameters 'order_by' and 'group_by'
        order_by_clause: Optional[Any] = None
        if self.ORDER_BY_PARAM in kwargs:
            if self.requires_order_by:
                raise excs.Error(
                    f'{self.display_name}(): order_by invalid, this function requires the first argument to be the '
                    f'ordering expression'
                )
            if not self.allows_window:
                raise excs.Error(
                    f'{self.display_name}(): order_by invalid with an aggregate function that does not allow windows'
                )
            order_by_clause = kwargs.pop(self.ORDER_BY_PARAM)
        elif self.requires_order_by:
            # the first argument is the order-by expr
            if len(args) == 0:
                raise excs.Error(f'{self.display_name}(): requires an ordering expression as its first argument')
            order_by_clause = args[0]
            if not isinstance(order_by_clause, exprs.Expr):
                raise excs.Error(
                    f'{self.display_name}(): the first argument needs to be a Pixeltable expression, but instead is a '
                    f'{type(order_by_clause)}'
                )
            # don't pass the first parameter on, the Function doesn't get to see it
            args = args[1:]

        group_by_clause: Optional[Any] = None
        if self.GROUP_BY_PARAM in kwargs:
            if not self.allows_window:
                raise excs.Error(
                    f'{self.display_name}(): group_by invalid with an aggregate function that does not allow windows'
                )
            group_by_clause = kwargs.pop(self.GROUP_BY_PARAM)

        args = [exprs.Expr.from_object(arg) for arg in args]
        kwargs = {k: exprs.Expr.from_object(v) for k, v in kwargs.items()}

        resolved_fn, bound_args = self._bind_to_matching_signature(args, kwargs)
        return_type = resolved_fn.call_return_type(bound_args)

        return exprs.FunctionCall(
            resolved_fn,
            args,
            kwargs,
            return_type,
            order_by_clause=[order_by_clause] if order_by_clause is not None else [],
            group_by_clause=[group_by_clause] if group_by_clause is not None else [],
        )

    def validate_call(self, bound_args: dict[str, 'exprs.Expr']) -> None:
        from pixeltable import exprs

        super().validate_call(bound_args)

        # check that init parameters are not Exprs
        # TODO: do this in the planner (check that init parameters are either constants or only refer to grouping exprs)
        for param_name in self.init_param_names[0]:
            if param_name in bound_args and not isinstance(bound_args[param_name], exprs.Literal):
                raise excs.Error(f'{self.display_name}(): init() parameter {param_name!r} must be a constant value')

    def __repr__(self) -> str:
        return f'<Pixeltable Aggregator {self.name}>'


# Decorator invoked without parentheses: @pxt.uda
@overload
def uda(decorated_fn: Callable) -> AggregateFunction: ...


# Decorator schema invoked with parentheses: @pxt.uda(**kwargs)
@overload
def uda(
    *,
    requires_order_by: bool = False,
    allows_std_agg: bool = True,
    allows_window: bool = False,
    type_substitutions: Optional[Sequence[dict]] = None,
) -> Callable[[type[Aggregator]], AggregateFunction]: ...


def uda(*args, **kwargs):  # type: ignore[no-untyped-def]
    """Decorator for user-defined aggregate functions.

    The decorated class must inherit from Aggregator and implement the following methods:
    - __init__(self, ...) to initialize the aggregator
    - update(self, ...) to update the aggregator with a new value
    - value(self) to return the final result

    The decorator creates an AggregateFunction instance from the class and adds it
    to the module where the class is defined.

    Parameters:
    - requires_order_by: if True, the first parameter to the function is the order-by expression
    - allows_std_agg: if True, the function can be used as a standard aggregate function w/o a window
    - allows_window: if True, the function can be used with a window
    """
    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        # Decorator invoked without parentheses: @pxt.uda
        # Simply call make_aggregator with defaults.
        return make_aggregator(cls=args[0])

    else:
        # Decorator schema invoked with parentheses: @pxt.uda(**kwargs)
        # Create a decorator for the specified schema.
        requires_order_by = kwargs.pop('requires_order_by', False)
        allows_std_agg = kwargs.pop('allows_std_agg', True)
        allows_window = kwargs.pop('allows_window', False)
        type_substitutions = kwargs.pop('type_substitutions', None)
        if len(kwargs) > 0:
            raise excs.Error(f'Invalid @uda decorator kwargs: {", ".join(kwargs.keys())}')
        if len(args) > 0:
            raise excs.Error('Unexpected @uda decorator arguments.')

        def decorator(cls: type[Aggregator]) -> AggregateFunction:
            return make_aggregator(
                cls,
                requires_order_by=requires_order_by,
                allows_std_agg=allows_std_agg,
                allows_window=allows_window,
                type_substitutions=type_substitutions,
            )

        return decorator


def make_aggregator(
    cls: type[Aggregator],
    requires_order_by: bool = False,
    allows_std_agg: bool = True,
    allows_window: bool = False,
    type_substitutions: Optional[Sequence[dict]] = None,
) -> AggregateFunction:
    class_path = f'{cls.__module__}.{cls.__qualname__}'
    instance = AggregateFunction(cls, type_substitutions, class_path, requires_order_by, allows_std_agg, allows_window)
    # do the path validation at the very end, in order to be able to write tests for the other failure cases
    validate_symbol_path(class_path)
    return instance



from __future__ import annotations

import abc
from typing import Any

import sqlalchemy as sql

from pixeltable import catalog, exprs


class IndexBase(abc.ABC):
    """
    Internal interface used by the catalog and runtime system to interact with indices:
    - types and expressions needed to create and populate the index value column
    - creating/dropping the index
    This doesn't cover querying the index, which is dependent on the index semantics and handled by
    the specific subclass.
    """

    @abc.abstractmethod
    def __init__(self, c: catalog.Column, **kwargs: Any):
        pass

    @abc.abstractmethod
    def index_value_expr(self) -> exprs.Expr:
        """Return expression that computes the value that goes into the index"""
        pass

    @abc.abstractmethod
    def records_value_errors(self) -> bool:
        """True if index_value_expr() can raise errors"""
        pass

    @abc.abstractmethod
    def index_sa_type(self) -> sql.types.TypeEngine:
        """Return the sqlalchemy type of the index value column"""
        pass

    @abc.abstractmethod
    def create_index(self, index_name: str, index_value_col: catalog.Column) -> None:
        """Create the index on the index value column"""
        pass

    @classmethod
    @abc.abstractmethod
    def display_name(cls) -> str:
        pass

    @abc.abstractmethod
    def as_dict(self) -> dict:
        pass

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, c: catalog.Column, d: dict) -> IndexBase:
        pass



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `AudioType`.

Example:
```python
import pixeltable as pxt
import pixeltable.functions as pxtf

t = pxt.get_table(...)
t.select(pxtf.audio.get_metadata()).collect()
```
"""

import pixeltable as pxt
from pixeltable.utils.code import local_public_names


@pxt.udf(is_method=True)
def get_metadata(audio: pxt.Audio) -> dict:
    """
    Gets various metadata associated with an audio file and returns it as a dictionary.
    """
    return pxt.functions.video._get_metadata(audio)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from typing import Iterable, Optional

import sqlalchemy as sql

from .expr import Expr
from .expr_dict import ExprDict


class SqlElementCache:
    """Cache of sql.ColumnElements for exprs"""

    cache: dict[int, Optional[sql.ColumnElement]]  # key: Expr.id

    def __init__(self, elements: Optional[ExprDict[sql.ColumnElement]] = None):
        self.cache = {}
        if elements is not None:
            for e, el in elements.items():
                self.cache[e.id] = el

    def extend(self, elements: ExprDict[sql.ColumnElement]) -> None:
        for e, el in elements.items():
            self.cache[e.id] = el

    def get(self, e: Expr) -> Optional[sql.ColumnElement]:
        """Returns the sql.ColumnElement for the given Expr, or None if Expr.to_sql() returns None."""
        try:
            return self.cache[e.id]
        except KeyError:
            pass
        el = e.sql_expr(self)
        self.cache[e.id] = el
        return el

    def contains(self, item: Expr) -> bool:
        """Returns True if the cache contains a (non-None) value for the given Expr."""
        return self.get(item) is not None

    def contains_all(self, items: Iterable[Expr]) -> bool:
        """Returns True if the cache contains a (non-None) value for every item in the collection of Exprs."""
        return all(self.get(e) is not None for e in items)



import PIL.Image

from pixeltable.env import Env


def resolve_torch_device(device: str, allow_mps: bool = True) -> str:
    Env.get().require_package('torch')
    import torch

    if device == 'auto':
        if torch.cuda.is_available():
            return 'cuda'
        if allow_mps and torch.backends.mps.is_available():
            return 'mps'
        return 'cpu'
    return device


def normalize_image_mode(image: PIL.Image.Image) -> PIL.Image.Image:
    """
    Converts grayscale images to 3-channel for compatibility with models that only work with
    multichannel input.
    """
    if image.mode in ('1', 'L'):
        return image.convert('RGB')
    if image.mode == 'LA':
        return image.convert('RGBA')
    return image



from __future__ import annotations

import dataclasses
import enum
import itertools
import logging
from typing import Optional

from typing_extensions import Self

import pixeltable.exceptions as excs

_logger = logging.getLogger('pixeltable')

# name of the position column in a component view
_POS_COLUMN_NAME = 'pos'
_ROWID_COLUMN_NAME = '_rowid'

# Set of symbols that are predefined in the `InsertableTable` class (and are therefore not allowed as column names).
# This will be populated lazily to avoid circular imports.
_PREDEF_SYMBOLS: Optional[set[str]] = None


@dataclasses.dataclass
class UpdateStatus:
    """
    Information about updates that resulted from a table operation.
    """

    num_rows: int = 0
    # TODO: disambiguate what this means: # of slots computed or # of columns computed?
    num_computed_values: int = 0
    num_excs: int = 0
    updated_cols: list[str] = dataclasses.field(default_factory=list)
    cols_with_excs: list[str] = dataclasses.field(default_factory=list)

    def __iadd__(self, other: 'UpdateStatus') -> Self:
        self.num_rows += other.num_rows
        self.num_computed_values += other.num_computed_values
        self.num_excs += other.num_excs
        self.updated_cols = list(dict.fromkeys(self.updated_cols + other.updated_cols))
        self.cols_with_excs = list(dict.fromkeys(self.cols_with_excs + other.cols_with_excs))
        return self

    @property
    def insert_msg(self) -> str:
        """Return a message describing the results of an insert operation."""
        if self.num_excs == 0:
            cols_with_excs_str = ''
        else:
            cols_with_excs_str = (
                f' across {len(self.cols_with_excs)} column{"" if len(self.cols_with_excs) == 1 else "s"}'
            )
            cols_with_excs_str += f' ({", ".join(self.cols_with_excs)})'
        msg = (
            f'Inserted {self.num_rows} row{"" if self.num_rows == 1 else "s"} '
            f'with {self.num_excs} error{"" if self.num_excs == 1 else "s"}{cols_with_excs_str}.'
        )
        return msg


class MediaValidation(enum.Enum):
    ON_READ = 0
    ON_WRITE = 1

    @classmethod
    def validated(cls, name: str, error_prefix: str) -> MediaValidation:
        try:
            return cls[name.upper()]
        except KeyError:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{error_prefix} must be one of: [{val_strs}]') from None


class IfExistsParam(enum.Enum):
    ERROR = 0
    IGNORE = 1
    REPLACE = 2
    REPLACE_FORCE = 3

    @classmethod
    def validated(cls, param_val: str, param_name: str) -> IfExistsParam:
        try:
            return cls[param_val.upper()]
        except KeyError:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{param_name} must be one of: [{val_strs}]') from None


class IfNotExistsParam(enum.Enum):
    ERROR = 0
    IGNORE = 1

    @classmethod
    def validated(cls, param_val: str, param_name: str) -> IfNotExistsParam:
        try:
            return cls[param_val.upper()]
        except KeyError:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{param_name} must be one of: [{val_strs}]') from None


def is_valid_identifier(name: str) -> bool:
    return name.isidentifier() and not name.startswith('_')


def is_valid_path(path: str, empty_is_valid: bool) -> bool:
    if not path:
        return empty_is_valid
    return all(is_valid_identifier(part) for part in path.split('.'))


def is_system_column_name(name: str) -> bool:
    from pixeltable.catalog import InsertableTable, View

    global _PREDEF_SYMBOLS  # noqa: PLW0603
    if _PREDEF_SYMBOLS is None:
        _PREDEF_SYMBOLS = set(itertools.chain(dir(InsertableTable), dir(View)))
    return name == _POS_COLUMN_NAME or name in _PREDEF_SYMBOLS
